# -*- coding: utf-8 -*-
"""

Fitting Tools embedded RADIS, for simple cases

For more advanced cases, use Fitroom : https://github.com/radis/fitroom

https://user-images.githubusercontent.com/16088743/120166810-4ac14980-c1fd-11eb-9dd5-8fb037db8793.mp4


Originally in radis-examples : https://github.com/radis/radis-examples

"""

import sys
from os.path import join
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import OptimizeResult, minimize

from radis import Spectrum, SpectrumFactory
from radis.phys.units import Unit
from radis.spectrum import plot_diff
from radis.spectrum.compare import get_default_units, get_diff, get_residual


# Calculate a new spectrum for given parameters:
def LTEModel(factory, fittable_parameters, fixed_parameters={}) -> Spectrum:
    """A model returning a single-slab LTE spectrum

    Parameters
    ----------
    model_input: dict
        input dictionary (typically: temperatures). Example
        ::
            {'Trot':value}

    Returns
    -------
    Spectrum: calculated spectrum

    Examples
    --------

    .. minigallery:: radis.tools.fitting.LTEModel

    See Also
    --------
    :py:func:`~radis.tools.fitting.Tvib12Tvib3Trot_NonLTEModel`
    :py:func:`~radis.tools.new_fitting.residual_LTE`

    """

    kwargs = {"name": "fit"}

    # ... create whatever model below (can have several slabs with SerialSlabs
    # ... or MergeSlabs, etc.)

    # update with remaining fittable and fixed parameters
    kwargs = {**kwargs, **fittable_parameters}
    kwargs = {**kwargs, **fixed_parameters}

    # >>> This is where the RADIS calculation is done!

    s = factory.eq_spectrum(**kwargs)

    # <<<

    # ... output should be a Spectrum object
    return s


# Calculate a new spectrum for given parameters:
def Tvib12Tvib3Trot_NonLTEModel(
    factory, fittable_parameters, fixed_parameters={}
) -> Spectrum:
    """A model returning a single-slab non-LTE spectrum with Tvib=(T12, T12, T3), Trot

    Used for non-LTE CO2 modeling. See the :ref:`multi-temperature fit example<example_multi_temperature_fit>`

    Parameters
    ----------
    fittable_parameters: dict
        input dictionary of fittable parameters (typically: temperatures). Example
        ::
            {'T12':value,
             'T3':value,
             'Trot':value}
    fixed_parameters: dict
        input dictionary of non-fittable parameters. Will override input
        conditions given to :py:meth:`~radis.lbl.factory.non_eq_spectrum`

    Returns
    -------
    Spectrum: calculated spectrum

    Examples
    --------

    .. minigallery:: radis.tools.fitting.Tvib12Tvib3Trot_NonLTEModel

    See Also
    --------
    :py:func:`~radis.tools.fitting.LTEModel`
    :py:func:`~radis.tools.new_fitting.residual_NonLTE`
    """

    fittable_parameters = fittable_parameters.copy()

    # ... input should remain a dict
    T12 = fittable_parameters.pop("T12")
    T3 = fittable_parameters.pop("T3")
    Trot = fittable_parameters.pop("Trot")

    # ... create whatever model below (can have several slabs with SerialSlabs
    # ... or MergeSlabs, etc.)

    # >>> This is where the RADIS calculation is done!

    kwargs = {
        "Tvib": (T12, T12, T3),
        "Trot": Trot,
        "Ttrans": Trot,
        # "vib_distribution":"treanor",
        "name": "fit",
    }
    # update with remaining fittable and fixed parameters
    kwargs = {**kwargs, **fittable_parameters}
    kwargs = {**kwargs, **fixed_parameters}

    s = factory.non_eq_spectrum(**kwargs)

    # <<<

    # ... output should be a Spectrum object
    return s


ite = 2


def fit_legacy(
    factory,
    s_exp,
    model,
    fit_parameters,
    bounds={},
    fixed_parameters={},
    plot=False,
    verbose=True,
    solver_options={"maxiter": 300},
) -> Union[Spectrum, OptimizeResult]:
    """Fit an experimental spectrum with an arbitrary model and an arbitrary
    number of fit parameters. This legacy function is replaced by :py:func:`~radis.tools.new_fitting.fit_spectrum`
    but is still functional.

    Parameters
    ----------
    s_exp : Spectrum
        experimental spectrum. Should have only spectral array only. Use
        :py:meth:`~radis.spectrum.spectrum.Spectrum.take`, e.g::
            sf.fit_legacy(s_exp.take('transmittance'))
    model : func -> Spectrum
        a line-of-sight model returning a Spectrum. Example :
        :py:func:`~radis.tools.fitting.LTEModel, `:py:func:`~radis.tools.fitting.Tvib12Tvib3Trot_NonLTEModel`
    fit_parameters : dict
        example::

            {fit_parameter:initial_value}s
    bounds : dict, optional
        example::

            {fit_parameter:[min, max]}
    fixed_parameters : dict
        fixed parameters given to the model. Example::

            fit_spectrum(fixed_parameters={"vib_distribution":"treanor"})

    Other Parameters
    ----------------
    plot: bool
        if ``True``, plot spectra as they are computed; and plot the convergence of
        the residual. Default ``False``
    verbose: bool, or ``2``
        use ``2`` for high verbose level.
    solver_options: dict
        parameters forwarded to the solver. More info in `~scipy.optimize.minimize`
        Example::

            {"maxiter": (int)  max number of iteration default ``300``,
             }

    Returns
    -------
    s_best: Spectrum
        best spectrum
    res: OptimizeResults
        output of `~scipy.optimize.minimize`

    Examples
    --------
    See a :ref:`one-temperature fit example <example_multi_temperature_fit>` or
    a :ref:`multi-temperature fit example <example_multi_temperature_fit>`.

    More advanced tools for interactive fitting of multi-dimensional, multi-slabs
    spectra can be found in :py:mod:`fitroom`.

    See Also
    --------
    :py:meth:`~radis.lbl.factory.SpectrumFactory.fit_legacy`,
    :py:func:`~radis.tools.new_fitting.fit_spectrum`,
    :py:mod:`fitroom`
    """

    # Get model being fitted:
    compute_los_model = lambda fit_parameters: model(
        factory, fit_parameters, fixed_parameters=fixed_parameters
    )

    # Calculate initial Spectrum, by showing all steps.
    compute_los_model(
        fit_parameters
    )  # Blank run to load energies; initialize all caches, etc.
    default_verbose = factory.verbose
    factory.verbose = 0  # reduce verbose during calculation.
    s0 = compute_los_model(fit_parameters)  # New run to get performance profile of fit
    sys.stderr.flush()
    s0.name = "Fit (in progress)"
    if "profiler" in s0.conditions and verbose > 0:
        print("-" * 30)
        print("TYPICAL FIT CALCULATION TIME:")
        s0.print_perf_profile()
        print("-" * 30)

    # %% Leastsq version

    # %%
    # User Params
    # -----------

    fit_params = list(fit_parameters.keys())
    bounds_arr = np.array([bounds[k] for k in fit_params])
    # fit_units = ['K', 'K', 'K']
    import astropy.units as u

    fit_units = []
    for k in fit_params:
        if isinstance(fit_parameters[k], u.Quantity) or isinstance(
            fit_parameters[k], u.Unit
        ):
            fit_units.append(Unit(fit_parameters[k]))
        else:
            fit_units.append("")

    if len(s_exp.get_vars()) != 1:
        raise ValueError(
            "More than one spectral array in experimental spectrum"
            + f"({len(s_exp.get_vars())}) : {s_exp.get_vars()}. "
            + "Choose one only with `s_exp.take('your_variable')`"
        )
    fit_variable = s_exp.get_vars()[0]

    # -----------------------------------------------------------------------------
    # -----------------------------------------------------------------------------
    #                 FITTING MACHINERY    (you shouldnt need to edit this)
    #                  ... just a few functions to make nice plots along
    #                  ... the fitting procedure
    # -----------------------------------------------------------------------------
    # -----------------------------------------------------------------------------

    fit_variable, wunit, Iunit = get_default_units(s_exp, s0, var=fit_variable)

    # %%
    # Algorithm Params
    # ----------------

    history_x = []
    history_res = []

    def print_fit_values(fit_values):
        return ",".join(
            [
                "{0}={1}{2}".format(
                    fit_params[i], np.round(fit_values[i], 0), fit_units[i]
                )
                for i in range(len(fit_params))
            ]
        )

    def generate_spectrum(fit_values) -> Spectrum:

        # Generate dictionary
        inputs = fit_parameters.copy()
        for k, v in zip(fit_params, fit_values):
            inputs[k] = v

        # Calculate the theoretical model
        s = compute_los_model(inputs)

        return s

    blit = True

    if plot:
        # plt.ion()
        # Graph with plot diff
        # figSpec, axSpec = plt.subplots(num='diffspectra')
        figSpec, axSpec = plot_diff(
            s_exp, s0, fit_variable, nfig="diffspectra", wunit=wunit, Iunit=Iunit
        )
        lineSpec = axSpec[0].get_lines()[1]
        lineDiff = axSpec[1].get_lines()[0]
        figSpec.canvas.draw()

        if blit:
            # cache the background
            # ... remove data first:
            s_diff = get_diff(s_exp, s0, var=fit_variable, wunit=wunit, Iunit=Iunit)
            lineSpec.set_data(s0.get(fit_variable)[0], s0.get(fit_variable)[1] * np.nan)
            lineDiff.set_data(s_diff[0], s_diff[1] * np.nan)
            figSpec.canvas.draw()
            # ... save:
            axSpec0background = figSpec.canvas.copy_from_bbox(axSpec[0].bbox)
            axSpec1background = figSpec.canvas.copy_from_bbox(axSpec[1].bbox)
            # ... re-add :
            lineSpec.set_data(s0.get(fit_variable))
            lineDiff.set_data(s_diff)
            figSpec.canvas.draw()

        plt.show(block=False)

    def cost_function(fit_values, plot=None):
        r"""Return error on Spectrum s vs experimental spectrum"""

        s = generate_spectrum(fit_values)

        # Delete unnecessary variables (for a faster resampling)
        for var in [k for k in s._q.keys() if k not in [fit_variable, "wavespace"]]:
            del s._q[var]

        if plot:  #  plot difference
            s_diff = get_diff(s_exp, s, var=fit_variable, wunit=wunit, Iunit=Iunit)
            lineSpec.set_data(s.get(fit_variable, wunit=wunit, Iunit=Iunit))
            lineDiff.set_data(s_diff)
            # axSpec[0].set_title(print_fit_values(fit_values))
            # plt.show(block=False)

            if blit:
                # ... from https://stackoverflow.com/questions/40126176/fast-live-plotting-in-matplotlib-pyplot
                # restore background
                figSpec.canvas.restore_region(axSpec0background)
                figSpec.canvas.restore_region(axSpec1background)

                # redraw just the points
                axSpec[0].draw_artist(lineSpec)
                axSpec[1].draw_artist(lineDiff)
                # ax2.draw_artist(text) # TODO

                # fill in the axes rectangle
                figSpec.canvas.blit(axSpec[0].bbox)
                figSpec.canvas.blit(axSpec[1].bbox)

            else:
                # redraw everything
                figSpec.canvas.draw()

            figSpec.canvas.flush_events()

        s.resample(s_exp, energy_threshold=2e-2)

        return get_residual(s, s_exp, fit_variable, ignore_nan=True, norm="L2")

    def log_cost_function(fit_values, plot=None):
        r"""Calls the cost_function, and write the values to the Log history"""

        res = cost_function(fit_values, plot=plot)

        history_x.append(fit_values)
        history_res.append(res)

        return res

    # Graph with residual
    # ... unlike 1D we cant plot the temperature here. Just plot the iteration

    if plot:
        plt.close("residual")
        figRes, axRes = plt.subplots(num="residual", figsize=(13.25, 6))
        axValues = axRes.twinx()

    fit_values_min, fit_values_max = bounds_arr.T
    res0 = log_cost_function(fit_values_min, plot=plot)
    res1 = log_cost_function(fit_values_max, plot=plot)

    maxiter = solver_options.get("maxiter", 300)

    if plot:
        # we need to plot lineValues already to get the legend right:
        lineValues = {}
        for i, k in enumerate(fit_params):
            lineValues[k] = axValues.plot(
                (1, 2), (fit_values_min[i], fit_values_max[i]), "-", label=k
            )[0]

        axRes.set_xlim((0, int(max(1, maxiter * 4))))
        axRes.set_ylim(ymin=0, ymax=1.5 * max(res0, res1))
        axValues.set_ylim(ymin=0, ymax=max(fit_values_max) * 1.1)
        axRes.set_xlabel("Iteration")
        axRes.set_ylabel("Residual")
        figRes.legend(loc="upper right")
        figRes.canvas.draw()
        if blit:
            axResbackground = figRes.canvas.copy_from_bbox(axRes.bbox)
        plt.show(block=False)

        (lineRes,) = axRes.plot((1, 2), (res0, res1), "-ko")
        (lineLast,) = axRes.plot(2, res0, "or")  # last iteration in red

    factory.verbose = False
    factory.warnings["NegativeEnergiesWarning"] = "ignore"

    global ite
    ite = 2

    if plot:
        plot_every = max(
            1, int(0.2 / s0.conditions["calculation_time"])
        )  # refresh plot every X calculations

    def cost_and_plot_function(fit_values):
        r"""Return error on Spectrum s vs experimental spectrum

        This is the function that is called by minimize()"""
        global ite
        ite += 1
        # Plot one spectrum every X ites
        plot_ite = plot and not ite % plot_every

        res = log_cost_function(fit_values, plot=plot_ite)

        if plot:
            # Add to plot history
            x, y = lineRes.get_data()
            lineRes.set_data((np.hstack((x, ite)), np.hstack((y, res))))
            # Add values to history
            for k, v in zip(fit_params, fit_values):
                x, y = lineValues[k].get_data()
                lineValues[k].set_data((np.hstack((x, ite)), np.hstack((y, v))))
            # Plot last
            lineLast.set_data([ite], [res])

            if blit:
                # ... from https://stackoverflow.com/questions/40126176/fast-live-plotting-in-matplotlib-pyplot
                figRes.canvas.restore_region(axResbackground)
                for k in fit_params:
                    axRes.draw_artist(lineValues[k])
                axRes.draw_artist(lineLast)
                axRes.draw_artist(lineRes)
                # fill in the axes rectangle
                figRes.canvas.blit(axRes.bbox)
            else:
                figRes.canvas.draw()
            figRes.canvas.flush_events()
            # plt.show(block=False)

        if verbose:
            print(
                "{0}, Residual: {1:.4f} {2}".format(
                    print_fit_values(fit_values),
                    res,
                    " 🏆" if res == min(history_res) else "",
                ),
                flush=True,
            )

        return res

    # %%>>> This is where the fitting loop happens
    if verbose:
        print("\nNow starting the fitting process:")
    if verbose:
        print("---------------------------------\n")
    solver_options = solver_options.copy()
    best = minimize(
        cost_and_plot_function,
        (fit_values_max + fit_values_min) / 2,
        method=solver_options.pop("method", None),
        jac=None,
        bounds=bounds_arr,
        options={
            **{
                "maxiter": maxiter,  # somehow.
                "eps": 20,
                # 'ftol':1e-10,
                # 'gtol':1e-10,
                "disp": verbose,
            },
            **solver_options,
        },
    )

    # %% Get best :

    s_best = generate_spectrum(best.x)

    if verbose and best.success:
        print(
            "Init {0} = {1}{2}".format(
                fit_params, (fit_values_max + fit_values_min) / 2, fit_units
            )
        )
        print("Final {0} = {1}{2}".format(fit_params, np.round(best.x), fit_units))

    if verbose >= 2:
        print(best)

    # Res history

    # ... what does history say:
    if verbose:
        print(
            "Best {0} = {1}{2} reached at iteration {3}/{4}".format(
                fit_params,
                history_x[np.argmin(history_res)],
                fit_units,
                np.argmin(history_res),
                best.nfev,
            )
        )

    # ... note that there are more function evaluations (best.nfev) that actual solver
    # ... iterations (best.nit) because the Jacobian is calculated numerically with
    # ... internal function calls

    # Close
    factory.verbose = default_verbose

    return s_best, best


if __name__ == "__main__":

    # %% Get Fitted Data
    from radis.test.utils import getValidationCase, setup_test_line_databases

    setup_test_line_databases()

    # Data from Dang, adapted by Klarenaar, digitized by us
    s_exp = Spectrum.from_txt(
        getValidationCase(
            join(
                "test_CO2_3Tvib_vs_klarenaar_data", "klarenaar_2017_digitized_data.csv"
            )
        ),
        "transmittance_noslit",
        wunit="cm-1",
        unit="",
        delimiter=",",
        name="Klarenaar 2017",
    )

    # %% Calculate

    sf = SpectrumFactory(
        2284.2,
        2284.6,
        wstep=0.001,  # cm-1
        pressure=20 * 1e-3,  # bar
        cutoff=1e-25,
        isotope="1,2",
        path_length=10,  # cm-1
        mole_fraction=0.1 * 28.97 / 44.07,
        truncation=0.5,  # cm-1
        neighbour_lines=0,
        medium="vacuum",
        # parsum_mode="tabulation"
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITEMP-CO2-TEST")

    s_best, best = sf.fit_legacy(
        s_exp.take("transmittance_noslit"),
        model=Tvib12Tvib3Trot_NonLTEModel,
        fit_parameters={
            "T12": 517,
            "T3": 2641,
            "Trot": 491,
        },
        bounds={"T12": [300, 2000], "T3": [300, 5000], "Trot": [300, 2000]},
        fixed_parameters={"vib_distribution": "treanor"},
        plot=True,
        solver_options={"method": "TNC", "maxiter": 200},
    )
    plot_diff(s_exp, s_best)

    s_best.print_perf_profile()



# -*- coding: utf-8 -*-
"""

-------------------------------------------------------------------------------


"""


def convert_and_strip_units(quantity, output_unit=None, digit=10):
    """Strips units and return the numerical value.

    Parameters
    ----------
    quantity : int or float or list or None or `~astropy.units.quantity.Quantity`
        Numerical quantity. Pass it without including units if default units are intended.
    output_unit : `~astropy.units.core.UnitBase` or `~astropy.units.quantity.Quantity`
        The default units in which the quantity is to be converted before extracting the value.

    Other Parameters
    ----------------

    round: int
        round to that number of digit after conversion. Default ``10``. Set to
        ``None`` to disable.

    Returns
    -------
    int or float or None or ~numpy.ndarray
        The numerical value extracted from ``quantity``

    Examples
    --------

        >>> convert_and_strip_units(4.5 * u.um, u.nm)
        <<< 4500.0

        >>> convert_and_strip_units(200000 * (u.m)**-1, 1/u.cm)
        <<< 2000.0

    Raises
    ------
    TypeError
        Raised when ``quantity`` is a astropy.units quantity and ``output_unit`` is ``None``.
    """

    import astropy.units as u

    if isinstance(quantity, u.Quantity):
        from astropy.units.imperial import deg_F

        # TODO : make it possible to test if not adimensioned, before loading.
        # This would allow not to have to load Astropy.units on start-up?
        if output_unit in (u.deg_C, deg_F, u.K):
            quantity = quantity.to_value(output_unit, equivalencies=u.temperature())
        elif isinstance(output_unit, (u.UnitBase, u.Quantity)):
            quantity = quantity.to_value(output_unit, equivalencies=u.spectral())
        else:
            raise TypeError(
                "'output_unit' parameter is not a valid astropy unit: {0}".format(
                    output_unit
                )
            )

        if digit:
            quantity = round(quantity, digit)

    return quantity


if __name__ == "__main__":
    import pytest

    pytest.main(["../test/phys/test_units.py"])



from ctypes import Structure, c_float, c_int, c_size_t


class threadIdx_t(Structure):
    _fields_ = [
        ("x", c_size_t),
        ("y", c_size_t),
        ("z", c_size_t),
    ]


class blockDim_t(Structure):
    _fields_ = [
        ("x", c_size_t),
        ("y", c_size_t),
        ("z", c_size_t),
    ]


class blockIdx_t(Structure):
    _fields_ = [
        ("x", c_size_t),
        ("y", c_size_t),
        ("z", c_size_t),
    ]


class gridDim_t(Structure):
    _fields_ = [
        ("x", c_size_t),
        ("y", c_size_t),
        ("z", c_size_t),
    ]


class initData_t(Structure):
    _fields_ = [
        ("v_min", c_float),
        ("dv", c_float),
        ("N_v", c_int),
        ("N_v_FT", c_int),
        ("N_x_FT", c_int),
        ("dxG", c_float),
        ("dxL", c_float),
        ("N_lines", c_int),
        ("N_collision_partners", c_int),
        ("log_c2Mm", c_float * 16),
    ]


class iterData_t(Structure):
    _fields_ = [
        ("p", c_float),
        ("log_2p", c_float),
        ("hlog_T", c_float),
        ("log_rT", c_float),
        ("c2T", c_float),
        ("N", c_float),
        ("x", c_float * 16),
        ("l", c_float),
        ("slit_FWHM", c_float),
        ("log_wG_min", c_float),
        ("log_wL_min", c_float),
        ("N_G", c_int),
        ("N_L", c_int),
        ("Q", c_float * 16),
    ]



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 14 15:38:18 2018

@author: erwan
"""


import pytest

from radis.misc.config import get_config, printDatabankList


@pytest.mark.fast
def test_json_config_file(*args, **kwargs):
    """Test that it's readable"""
    get_config()


@pytest.mark.fast
def test_databanks(*args, **kwargs):
    """Test radis.json is readable"""
    printDatabankList()


def _run_testcases(verbose=True, *args, **kwargs):
    """Test array functions"""

    test_json_config_file()
    test_databanks()

    return True


if __name__ == "__main__":
    print("test config files:", _run_testcases())



# -*- coding: utf-8 -*-
"""Created on Tue Jul 18 16:15:03 2017.

@author: erwan

Summary
-------

Define :py:class:`~radis.db.classes.Molecule`, :py:class:`~radis.db.classes.Isotope`
and :py:class:`~radis.db.classes.ElectronicState` classes

ElectronicState has:

- Molecular parameters (read from HITRAN molparams.txt)
- Vibrational and rotational energies

Calculate Vibrational and Rotational energies for some database molecules

Molecules
-----------

Shortcut to some molecules ground states are defined at the end of the file
in the dictionary `Molecules`, with the following structure:

{molecule_name: {isotope_number: {electronic_state_name: ElectronicState object}}}

Import with::

    from radis.db.molecules import Molecules
    CO2_X = Molecules['CO2'][1]['X']            # 1 for first isotope

Or::

    from radis import getMolecule                # directly
    CO2_2 = getMolecule('CO2', 1, 'X')         # better KeyError messages


Routine Listing
---------------
- :func:`~radis.db.classes.is_atom`
- :func:`~radis.db.classes.is_neutral`
- :func:`~radis.db.classes.roman_to_int`
- :func:`~radis.db.classes.int_to_roman`
- :func:`~radis.db.classes.get_ielem_charge`
- :func:`~radis.db.classes.to_conventional_name`
- :func:`~radis.db.classes.get_molecule`
- :func:`~radis.db.classes.get_molecule_identifier`

-------------------------------------------------------------------------------
"""

import re
from os.path import dirname, exists, join

import periodictable

from radis.db.conventions import get_convention
from radis.db.utils import (
    get_default_jsonfile,
    get_dunham_coefficients,
    get_herzberg_coefficients,
    parse_doi,
)
from radis.levels.dunham import EvJ, Fv, Gv

# %% Hitran groups and classes
# As defined in Rothman et al, "The HITRAN 2004 molecular spectroscopic database"
# Tables 3 and 4
# Groups and classes are needed to compute energy in nonequilibrium mode.

# Groups (define local quanta)

HITRAN_GROUP1 = [
    "H2O",
    "O3",
    "SO2",
    "NO2",
    "HNO3",
    "H2CO",
    "HOCl",
    "H2O2",
    "COF2",
    "H2S",
    "HO2",
    "HCOOH",
    "ClONO2",
    "HOBr",
    "C2H4",
]
"""str: asymmetric rotors"""

HITRAN_GROUP2 = [
    "CO2",
    "N2O",
    "CO",
    "HF",
    "HCl",
    "HBr",
    "HI",
    "OCS",
    "N2",
    "HCN",
    "C2H2",
    "NO+",
]
"""str: diatomic and linear molecules"""

HITRAN_GROUP3 = ["SF6", "CH4"]
"""str: Spherical rotors"""

HITRAN_GROUP4 = ["CH3D", "CH3Cl", "C2H6", "NH3", "PH3", "CH3OH"]
"""str: symmetric rotors"""

HITRAN_GROUP5 = ["O2"]
"""str: Triplet-Sigma ground electronic states"""

HITRAN_GROUP6 = ["NO", "OH", "ClO"]
"""str: Doublet-Pi ground electronic states"""

# Classes (define global quanta)

HITRAN_CLASS1 = ["CO", "HF", "HCl", "HBr", "HI", "N2", "NO+"]
"""str: Diatomic molecules with ? """

HITRAN_CLASS2 = ["O2"]
"""str: Diatomic molecules with different electronic levels"""

HITRAN_CLASS3 = ["NO", "OH", "ClO"]
"""str: Diatomic molecules with doublet-Pi electronic state"""

HITRAN_CLASS4 = ["N2O", "OCS", "HCN"]
"""str: Linear triatomic"""

HITRAN_CLASS5 = ["CO2"]
"""str: Linear triatomic with large Fermi resonance"""

HITRAN_CLASS6 = ["H2O", "O3", "SO2", "NO2", "HOCl", "H2S", "HO2", "HOBr"]
"""str: Non-linear triatomic"""

HITRAN_CLASS7 = ["C2H2"]
"""str: Linear tetratomic"""

HITRAN_CLASS8 = ["NH3", "PH3"]
"""str: Pyramidal tetratomic"""

HITRAN_CLASS9 = ["H2CO", "H2O2", "COF2"]
"""str: Non-linear tetratomic"""

HITRAN_CLASS10 = [
    "CH4",
    "CH3Cl",
    "C2H6",
    "HNO3",
    "SF6",
    "HCOOH",
    "ClONO2",
    "C2H4",
    "CH3OH",
]
"""str: Pentatomic or greater polyatomic"""
# 0.9.22: Removed CH3D has no HITRAN identifier


# %% HITRAN ids

trans = {
    "1": "H2O",
    "2": "CO2",
    "3": "O3",
    "4": "N2O",
    "5": "CO",
    "6": "CH4",
    "7": "O2",
    "8": "NO",
    "9": "SO2",
    "10": "NO2",
    "11": "NH3",
    "12": "HNO3",
    "13": "OH",
    "14": "HF",
    "15": "HCl",
    "16": "HBr",
    "17": "HI",
    "18": "ClO",
    "19": "OCS",
    "20": "H2CO",
    "21": "HOCl",
    "22": "N2",
    "23": "HCN",
    "24": "CH3Cl",
    "25": "H2O2",
    "26": "C2H2",
    "27": "C2H6",
    "28": "PH3",
    "29": "COF2",
    "30": "SF6",
    "31": "H2S",
    "32": "HCOOH",
    "33": "HO2",
    "34": "O",
    "35": "ClONO2",
    "36": "NO+",
    "37": "HOBr",
    "38": "C2H4",
    "39": "CH3OH",
    "40": "CH3Br",
    "41": "CH3CN",
    "42": "CF4",
    "43": "C4H2",
    "44": "HC3N",
    "45": "H2",
    "46": "CS",
    "47": "SO3",
    "48": "C2N2",
    "49": "COCl2",
    "50": "SO",
    "51": "CH3F",
    "52": "GeH4",
    "53": "CS2",
    "54": "CH3I",
    "55": "NF3",
}
HITRAN_MOLECULES = list(trans.values())
""" str: list of [HITRAN-2020]_ molecules. """


def is_atom(species):
    """
    determine whether the input species is atomic
    """
    return species.split("_")[0] in [
        i.symbol for i in list(periodictable.elements)[1:]
    ]  # first element is neutron, symbol 'n'


def is_neutral(species):
    """
    determine whether the input species is a neutral atom
    """
    return species.endswith("_I")


roman_numerals = {"I": 1, "V": 5, "X": 10, "L": 50, "C": 100, "D": 500, "M": 1000}


def roman_to_int(roman):
    """
    A function to convert a `str` of roman numerals to an `int`

    Should support 'Standard form', 'Other additive forms', and 'Other subtractive forms' mentioned here: https://en.wikipedia.org/wiki/Roman_numerals
    """
    roman = roman.upper()
    total = 0
    current_max = 0  # the current largest 'digit'
    for i in range(len(roman) - 1, -1, -1):  # in reverse, from last character to first
        current_max = max(roman_numerals[roman[i]], current_max)
        if roman_numerals[roman[i]] < current_max:
            total -= roman_numerals[roman[i]]
        else:
            total += roman_numerals[roman[i]]
    return total


invert_roman_numerals = dict(
    (v, k) for k, v in sorted(roman_numerals.items(), key=lambda x: x[1], reverse=True)
)


def int_to_roman(num):
    """
    A function to convert an `int` to a `str` of roman numerals
    """
    count = {}
    for factor in invert_roman_numerals:
        count[factor] = num // factor
        num = num % factor
    string = count[1000] * "M"
    nums = list(invert_roman_numerals)
    i = 0
    while i < len(invert_roman_numerals) - 2:
        if count[nums[i + 2]] == 4:
            if count[nums[i + 1]] == 1:  # so overall i+2 digit is 9
                addstring = (
                    invert_roman_numerals[nums[i + 2]] + invert_roman_numerals[nums[i]]
                )
            else:  # overall i+2 digit is 4
                addstring = (
                    invert_roman_numerals[nums[i + 2]]
                    + invert_roman_numerals[nums[i + 1]]
                )
        else:
            addstring = (
                count[nums[i + 1]] * invert_roman_numerals[nums[i + 1]]
                + count[nums[i + 2]] * invert_roman_numerals[nums[i + 2]]
            )
        string += addstring
        i += 2
    return string


def get_ielem_charge(species):
    """return the atomic number xx and the charge yy of a positive or neutral atomic species input in the form of a `str` 'xx.yy'"""
    ielemstr, chargem1str = species.split(".")
    ielem = int(ielemstr)
    charge = int(chargem1str)
    return ielem, charge


def to_conventional_name(species):
    """
    To convert an atomic species given in any of the following forms to spectroscopic notation (with an underscore rather than whitespace):

    - as a ``str`` in spectroscopic notation with a space separating the symbol and ionisation level (e.g. 'O II' for singly ionised oxygen)

    - as an ``int`` composed of the atomic number xx multiplied by 100 and then added to the charge (801 for O_II, 2600 for Fe_I)

    - as a **``str``** (not a float) in the form 'xx.yy', i.e. the charge divided by 100 and then added to the atomic number

    - For neutrals, as a ``str`` with just the atomic symbol

    - For ions, as a ``str`` composed of the atomic symbol followed by a string of + signs the number of which matches the charge on the ion

    If given a molecule name or HITRAN ID, it is returned unchanged
    """
    if isinstance(species, float):
        raise Exception(
            "If you are trying to input the species in the form xx.yy, input as a string, not a float"
        )
    if isinstance(species, int):
        atomic_number = species // 100
        if (
            atomic_number == 0
        ):  # looks like it's just the molecule ID of a molecule instead
            return species
        symbol = periodictable.elements[atomic_number]
        roman = int_to_roman(species % 100 + 1)
        return f"{symbol}_{roman}"
    if "." in species:
        ielem, charge = get_ielem_charge(species)
        symbol = periodictable.elements[ielem]
        roman = int_to_roman(charge + 1)
        return f"{symbol}_{roman}"
    if "+" in species:
        # Count the number of positive charges
        chargeplus1 = (
            species.count("+") + 1
        )  # Add one to convert to conventional notation
        # Get the element of the species without the charge
        element_name = species.split("+")[0]
        # Convert the charge to roman notation
        chargeplus1_in_roman = int_to_roman(chargeplus1)
        return f"{element_name}_{chargeplus1_in_roman}"
    if " " in species.strip():
        species = species.replace(" ", "_")
    atomic_symbol = species.split("_")[0]
    if atomic_symbol == species:
        try:  # confirm it actually is an atom
            getattr(periodictable, atomic_symbol)
        except AttributeError:
            return species
        else:
            return f"{species}_I"
    return species


def get_molecule_identifier(molecule_name):
    r"""
    For a given input molecular formula, return the corresponding
    :py:data:`~radis.db.classes.HITRAN_MOLECULES` identifier number [1]_.

    Parameters
    ----------
    molecular_formula : str
        The string describing the molecule.

    Returns
    -------
    M: int
        The HITRAN molecular identified number.

    References
    ----------
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    Function is from https://github.com/nzhagen/hitran/blob/master/hitran.py

    """

    # Invert the dictionary.
    trans2 = {v: k for k, v in trans.items()}

    try:
        return int(trans2[molecule_name])
    except KeyError:
        raise NotImplementedError(
            "Molecule '{0}' not supported. Choose one of {1}".format(
                molecule_name, sorted(list(trans2.keys()))
            )
        )


def get_molecule(molecule_id):
    r"""
    For a given input molecular identifier, return the corresponding
    :py:data:`~radis.db.classes.HITRAN_MOLECULES` name [1]_.

    Parameters
    ----------
    molecular_id: str
        Hitran identifier of the molecule.


    References
    ----------
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    """

    # assert str
    id = "{:d}".format(int(molecule_id))

    try:
        return trans[id]
    except KeyError:
        raise NotImplementedError(
            "Molecule ID '{0}' unknown. Choose one of {1}".format(molecule_id, trans)
        )


# %% ExoMol molecules


from radis.db.molparam import get_extra_molparams

isotope_full_names = get_extra_molparams()["isotope_full_name"]

EXOMOL_ONLY_ISOTOPES_NAMES = {
    (molecule, int(iso_number)): iso_name
    for molecule, iso_dict in isotope_full_names.items()
    for iso_number, iso_name in iso_dict.items()
}
"""All :py:data:`~radis.db.classes.HITRAN_MOLECULES` are also converted to their ExoMol full-name format
in :py:func:`~radis.io.exomol.get_exomol_full_isotope_name`
"""
assert EXOMOL_ONLY_ISOTOPES_NAMES[("FeH", 1)] == "56Fe-1H"
assert EXOMOL_ONLY_ISOTOPES_NAMES[("SiO", 1)] == "28Si-16O"
assert EXOMOL_ONLY_ISOTOPES_NAMES[("CN", 1)] == "12C-14N"


EXOMOL_MOLECULES = [
    "AlCl",
    "AlF",
    "AlH",
    "AlO",
    "AsH3",
    "BeH",
    "C2",
    "C3",
    "C2H2",
    "C2H4",
    "C3",
    "CaCl",
    "CH",
    "CH3",
    "CH3Cl",
    "CH3F",
    "CH4",
    "CH_p",
    "CN",
    "CO",
    "CO2",
    "CP",
    "CS",
    "CaF",
    "CaH",
    "CaO",
    "CaOH",
    "CrH",
    "FeH",
    "H2",
    "HBO",
    "H2CO",
    "H2CS",
    "H2O",
    "H2O2",
    "H2S",
    "H2_p",
    "H3O_p",
    "H3_p",
    "HBr",
    "HCN",
    "HCl",
    "HF",
    "HNO3",
    "HeH_p",
    # "K", #not anymore since 01/07/2024
    "KCl",
    "KF",
    "KOH",
    "LaO",
    "LiCl",
    "LiF",
    "LiH",
    "LiH_p",
    "LiOH",
    "MgF",
    "MgH",
    "MgO",
    "N2",
    "N2O",
    "NH",
    "NH3",
    "NO",
    "NS",
    # "Na",  #not anymore since 01/07/2024
    "NaCl",
    "NaF",
    "NaH",
    "NaO",
    "NaOH",
    "NiH",
    "O2",
    "OCS",
    "OH",
    "OH_p",
    "PF3",
    "PH",
    "PH3",
    "PN",
    "PO",
    "PS",
    "SH",
    "SO",  # new
    "SO2",
    "SO3",
    "ScH",
    "SiH",
    "SiH2",
    "SiH4",
    "SiN",
    "SiO",
    "SiO2",
    "SiS",
    "SO",
    "TiH",
    "TiO",
    "VO",
    "YO",
    "ZrO",
    "cis-P2H2",
    "trans-P2H2",
]
EXOMOL_ONLY_MOLECULES = sorted(
    list(set([M for M, iso in EXOMOL_ONLY_ISOTOPES_NAMES.keys()]))
)

for M in EXOMOL_ONLY_ISOTOPES_NAMES.keys():
    assert M[0] in EXOMOL_MOLECULES

# %% Molecule class


class Molecule(object):
    r"""Define a new molecule

    Parameters
    ----------
    name: str, int
        molecule name, or HITRAN identifier

    Other Parameters
    ----------------
    verbose: boolean
        more blabla

    See Also
    --------
    :py:class:`~radis.db.classes.Isotope`, :py:class:`~radis.db.classes.ElectronicState`,
    :py:func:`~radis.db.molecules.getMolecule`

    """

    def __init__(
        self, name, verbose=True
    ):  # @dev: no optional kwargs here (final stop)

        self.verbose = verbose

        # Get name and integer id
        if isinstance(name, str):
            self.name = name

            # Get name without parenthesis (without state) for HITRAN identification
            filtername = re.sub(r"[\(\[].*?[\)\]]", "", name)

            try:
                self.id = get_molecule_identifier(filtername)
            except NotImplementedError:  # Not an HITRAN molecule
                self.id = None
        elif type(name) == int:
            self.id = name
            self.name = get_molecule(name)
        else:
            raise ValueError("Wrong name type:", name)


# %% Isotope class


class Isotope(Molecule):
    r"""Create an isotope of a given :py:class:`~radis.db.classes.Molecule`

    Parameters
    ----------
    molecule_name: str, or int
        molecule name or HITRAN identifier
    isotope: int
        isotope identifier, sorted by decreasing abundance (cf HITRAN
        nomenclature)

    Other Parameters
    ----------------
    isotope_name: str
        (optional) isotope name. Default ''
    abundance: float
        isotopologue abundance. Default ``None``

    See Also
    --------
    :py:class:`~radis.db.classes.ElectronicState`, :py:class:`~radis.db.classes.Molecule`,
    :py:func:`~radis.db.molecules.getMolecule`
    """

    def __init__(
        self, molecule_name, isotope, isotope_name="", abundance=None, **kwargs
    ):

        super(Isotope, self).__init__(
            name=molecule_name, **kwargs
        )  # initialize Molecule

        # Check input
        if not isinstance(isotope, int):
            raise TypeError(
                "Wrong format for isotope {0}: expected int, got {1}".format(
                    isotope, type(isotope)
                )
            )

        # Store
        self.iso = isotope
        self.isotope_name = isotope_name
        self.abundance = abundance


# %% ElectronicState class

_term_symbols = {"Σ": "SIG", "Π": "PI", "Δ": "DEL"}


def _format_term_symbol(term_symbol):
    """standardized name of greek symbols."""
    for k, v in _term_symbols.items():
        term_symbol = term_symbol.replace(k, v)
    return term_symbol


def _print_term_symbol(term_symbol):
    """nice print of greek letters."""
    for k, v in _term_symbols.items():
        term_symbol = term_symbol.replace(v, k)
    return term_symbol


class ElectronicState(Isotope):
    r"""Define an electronic state of an :py:class:`~radis.db.classes.Isotope`
    of a :py:class:`~radis.db.classes.Molecule`

    Parameters
    ----------
    molecule_name: str, int
        molecule name, or HITRAN identifier
    isotope: int
        isotope number of the molecule (sorted by abundance on Earth, HITRAN
        convention).
    state: str
        Electronic state name
    term_symbol: str
        Term symbol. Default ''
    g_e: int
        Degeneracy. Default ``None``
    spectroscopic_constants: str, ``'default'``, or dict
        filename of spectroscopic constants under Herzberg or Dunham format.

        Expected in the file:

            Yij: cm-1
                rovibrational coefficients in Dunham convention

        or

            wexe, Be, etc. : cm-1
                rovibrational coefficients in Herzberg convention

        or

            Te: cm-1
                electronic energy. Default ``None`` if not given

        If ``default``, the constants defined in
        :ref:`spectroscopic constants <label_db_spectroscopic_constants>` are used.

        One can also directly give a dictionary directly::

            spectroscopic_constants = {'we':2300, 'wexe':20}

    spectroscopic_constants_type: ``'herzberg'``, ``'dunham'``
        convention for spectroscopic constants. Default ``'herzberg'``
    Erovib: function
        Rovibrational energy of the molecule. Accessible with, typically, Mol.Erovib(v,J)
    Ehaj: function
        To return vibrational energy in harmonic and anharmonic component,
        plus rotational energy. Used to build Treanor distributions
    vmax, vmax_morse: int, or ``None``
        maximum vibrational number (required for partition function calculation)
        to be calculated with Dunham expansion, and Morse potential
        If None, number will be inferred from dissociation energy
    Jmax: int, or ``None``
        maximum rotational number (required for partition function calculation)
        If None, number will be inferred from dissociation energy
    Ediss: cm-1
        dissociation energy. Required for partition function calculation
        if neither vmax nor Jmax are given
    kwargs: **dict
        forwarded to parent class

    Examples
    --------

    .. minigallery:: radis.db.classes.ElectronicState
        :add-heading:


    See Also
    --------
    :py:class:`~radis.db.classes.Isotope`, :py:class:`~radis.db.classes.Molecule`,
    :py:func:`~radis.db.molecules.getMolecule`

    """

    #'''
    # TODO
    # To allow for more flexibility, ElectronicState should also inherit from some
    # built-in Symmetry / species class, Sigma / Pi / Delta etc... where the energy
    # is calculated appropriately.  (ex: Rigid Rotor, Symmetric Top etc.)

    # (major code rewriting ahead...)
    # That should be part of RADIS extension to non Sigma electronic states.

    def __init__(
        self,
        molecule_name,
        isotope,
        state,
        term_symbol="",
        g_e=None,
        spectroscopic_constants="default",
        spectroscopic_constants_type="herzberg",
        Erovib=None,
        Ehaj=None,
        vmax=None,
        vmax_morse=None,
        Jmax=None,
        Ediss=None,
        **kwargs,
    ):

        super(ElectronicState, self).__init__(
            molecule_name=molecule_name, isotope=isotope, **kwargs
        )  # initialize Isotope

        # Initialize electronic state specific parameters
        self.state = state
        self.term_symbol = term_symbol
        self.g_e = g_e

        self.doi = (
            None  #: str: stores DOI of spectroscopic constants if given in database
        )
        self._parse_rovib_constants(
            spectroscopic_constants, spectroscopic_constants_type
        )

        #        self.Erovib = None   #: func: overwritten by Erovib if given, or by default Dunham developments if spectroscopic_constants are given
        self.Ehaj = None  #: func: overwritten by Erovib if given, or by default Dunham developments if spectroscopic_constants are given
        self._assign_E(Erovib, Ehaj)
        self.state = state
        self.vmax = vmax
        self.vmax_morse = vmax_morse
        self.Jmax = Jmax
        self.Ediss = Ediss

    def _parse_rovib_constants(
        self, spectroscopic_constants, spectroscopic_constants_type
    ):
        r"""Parse spectroscopic constants

        Stores :py:attr:`~radis.db.classes.ElectronicState.Te` and
        :py:attr:`~radis.db.classes.ElectronicState.re` as electronic state
        attributes, and the rest under :py:attr:`~radis.db.classes.ElectronicState.rovib_constants`

        Parameters
        ----------
        spectroscopic_constants: str, or ``'default'``, or dict
            filename of spectroscopic constants under Herzberg or Dunham format.

            Expected in the file:

                Yij: cm-1
                    rovibrational coefficients in Dunham convention

            or

                wexe, Be, etc. : cm-1
                    rovibrational coefficients in Herzberg convention

            or

                Te: cm-1
                    electronic energy. Default ``None`` if not given

            If ``default``, the constants defined in
            :ref:`spectroscopic constants <label_db_spectroscopic_constants>` are used.

            One can also directly give a dictionary directly::

                spectroscopic_constants = {'we':2300, 'wexe':20}

        spectroscopic_constants_type: ``'herzberg'``, ``'dunham'``
            convention for spectroscopic constants. Default ``'herzberg'``

        Returns
        -------
        None:
            but constants are stored under :py:attr:`~radis.db.classes.ElectronicState.rovib_constants`,
            and store json file in :py:attr:`~radis.db.classes.ElectronicState.jsonfile`

        """

        if isinstance(spectroscopic_constants, dict):
            rovib_constants = spectroscopic_constants
            jsonfile = None

            doi = parse_doi(rovib_constants)

        else:  # file

            # Get file name
            if spectroscopic_constants == "default":
                jsonfile = get_default_jsonfile(self.name)
            elif exists(spectroscopic_constants):  # absolute path
                jsonfile = spectroscopic_constants
            else:  # assume a json file stored in the default folder
                jsonfile = join(
                    dirname(get_default_jsonfile(self.name)), spectroscopic_constants
                )

            # Parse file
            if spectroscopic_constants_type == "dunham":
                rovib_constants, doi = get_dunham_coefficients(
                    self.name,
                    self.iso,
                    self.get_statename_utf(),
                    jsonfile=jsonfile,
                    return_doi=True,
                )
            elif spectroscopic_constants_type == "herzberg":
                rovib_constants, doi = get_herzberg_coefficients(
                    self.name,
                    self.iso,
                    self.get_statename_utf(),
                    jsonfile=jsonfile,
                    return_doi=True,
                )
            else:
                raise ValueError(
                    "Unexpected spectroscopic constant type: {0}".format(
                        spectroscopic_constants_type
                    )
                )

        # Clean keys
        # In particular, remove trailing '_cm-1' if given in dict or database
        import re

        rovib_constants = {
            re.sub("_cm-1$", "", k): v for (k, v) in rovib_constants.items()
        }

        # Get specific keys
        self.Te = rovib_constants.pop("Te", None)  # default None
        self.re = rovib_constants.pop("re", None)

        # Store
        self.rovib_constants = rovib_constants
        self.jsonfile = jsonfile
        self.doi = doi

    # Default method to calculate energy

    def _assign_E(self, Erovib, Ehaj):
        r"""Finds appropriate Electrorovibrational energy function for this molecule ,
        based on which convention is used for spectroscopic coefficients
        (Dunham, or Herzberg)

        Replaces the :py:meth:`~radis.db.classes.ElectronicState.Erovib` template
        with the appropriate function for this molecule / isotope / electronic state.

        Default methods only work for diatomic molecules. For polyatomic molecules,
        you should give the energy function directly (see how CO2 is defined in
        radis.db.molecules.py)

        Returns
        -------

        None:
            but assigns the ``self.Erovib`` function to the molecule

        """
        # Check input
        if Erovib is None and Ehaj is not None:
            raise ValueError(
                "If giving Ehaj (harmonic and anharmonic component calculation) "
                + "you must also supply Erovib (total rovibrational energy)"
            )

        if Erovib is not None:
            # override energy calculation
            self.Erovib = Erovib
            self.Ehaj = Ehaj
            if self.verbose >= 2:
                print(
                    "{0}: overriding Energy calculation with {1}".format(
                        self.get_fullname(), Erovib
                    )
                )
        else:
            # Autofind which energy model to use for the given molecule
            c = self.rovib_constants
            if len(c) == 0:
                pass  # keep the default function, that will raise an error on first call.
                if self.verbose >= 2:
                    print("{0}: No rovibconstants found".format(self.get_fullname()))
            else:
                convention = get_convention(c)  # Herzberg or Dunham

                if self.name in HITRAN_CLASS1:
                    if convention == "herzberg":
                        self.Erovib = self._E_Herzberg
                        self.Ehaj = None  # NotImplemented
                        if self.verbose >= 2:
                            print(
                                "{0}: using Herzberg coefficients for Energy calculation".format(
                                    self.get_fullname()
                                )
                            )
                    else:
                        self.Erovib = self._E_Dunham
                        self.Ehaj = None  # NotImplemented
                        if self.verbose >= 2:
                            print(
                                "{0}: using Dunham coefficients for Energy calculation".format(
                                    self.get_fullname()
                                )
                            )
                        # TODO: reformat these methods as external functions, but keep docstrings
                elif self.name in HITRAN_CLASS5:
                    if convention == "herzberg":
                        from radis.levels.energies_co2 import EvJ_co2, EvJah_co2

                        self._Erovib = EvJ_co2
                        self._Ehaj = EvJah_co2
                        self.Erovib = self._Erovib_default_coeffs
                        #                        self.Erovib.__doc__ == EvJ_co2.__doc__
                        self.Ehaj = self._Ehaj_default_coeffs
                    #                        self.Ehaj.__doc__ == EvJah_co2.__doc__
                    else:
                        raise NotImplementedError(
                            "Only Herzberg convention spectroscopic "
                            + "constants defined for {0}. ".format(self.name)
                            + "You still define your own Erovib function "
                            + "and overwrite the ElectronicState class one."
                        )

    def Erovib(self, v, J, remove_ZPE=True):
        r"""Calculate rovibrational energy of molecule

        .. math::

            E(v,J) = G(v) + F(v,J)

        Works with Herzberg convention for the spectroscopic coefficients used
        in :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv` ::

        Parameters
        ----------

        v: int
            vibrational state
            For polyatomic molecules, use ``v1, v2, v3, etc.`` instead of ``v``.

        J: int
            rotational state

        remove_ZPE: boolean
            if ``True``, removes energy of v=0,J=0 vibrational level (zero-point-energy)
            Default ``True``

        Returns
        -------

        energy of state in cm-1

        Notes
        -----

        Method is overwritten on molecule creation for polyatomic molecules.
        Refer to the correct method below:

        - CO2: :py:func:`~radis.levels.energies_co2.EvJ_co2`


        See Also
        --------

        :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv`


        """

        raise NotImplementedError(
            "Rovibrational energy not implemented for {0}".format(self.get_fullname)
        )  # @dev: see radis.db.classes.ElectronicState._assign_E

    def _Erovib_default_coeffs(self, *args, **kwargs):
        r"""Calculate rovibrational energy of molecule

        .. math::

            E(v,J) = G(v) + F(v,J)

        Works with Herzberg convention for the spectroscopic coefficients used
        in :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv` ::

        Parameters
        ----------

        v: int
            vibrational state
            For polyatomic molecules, use ``v1, v2, v3, etc.`` instead of ``v``.

        J: int
            rotational state

        remove_ZPE: boolean
            if ``True``, removes energy of v=0,J=0 vibrational level (zero-point-energy)
            Default ``True``

        Returns
        -------

        energy of state in cm-1

        Notes
        -----

        Method is overwritten on molecule creation for polyatomic molecules.
        Refer to the correct method below:

        - CO2: :py:func:`~radis.levels.energies_co2.EvJ_co2`


        See Also
        --------

        :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv`

        """
        assert "coeff_dict" not in kwargs
        kwargs.update({"coeff_dict": self.rovib_constants})

        return self._Erovib(*args, **kwargs)

    def _Ehaj_default_coeffs(self, *args, **kwargs):
        r"""Call default _Ehaj() function with default rovib_constants
        coefficients."""

        assert "coeff_dict" not in kwargs
        kwargs.update({"coeff_dict": self.rovib_constants})

        return self._Ehaj(*args, **kwargs)

    def _E_Dunham(self, v, J, remove_ZPE=True):
        # TODO: move in levels.energies as was done for CO2

        c = self.rovib_constants

        if remove_ZPE:
            ZPE = EvJ(0, 0, **c)
        else:
            ZPE = 0

        return EvJ(v, J, **c) - ZPE

    def _E_Herzberg(self, v, J, remove_ZPE=True):
        # TODO: move in levels.energies as was done for CO2
        r"""Calculate rovibrational energy of molecule

        .. math::

            E(v,J) = G(v) + F(v,J)

        Works with Herzberg convention for the spectroscopic coefficients used
        in :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv` ::

        Only works for diatomic molecules. Method should be overwritten on molecule
        creation for other molecules

        Parameters
        ----------

        v: int
            vibrational state

        J: int
            rotational state

        remove_ZPE: boolean
            if ``True``, removes energy of v=0,J=0 vibrational level (zero-point-energy)
            Default ``True``

        Returns
        -------

        energy of state in cm-1

        See Also
        --------

        :func:`~radis.levels.dunham.Gv`, :func:`~radis.levels.dunham.Fv`

        """

        try:

            # Get from constants (hardcoded, see reference above)
            c = self.rovib_constants

            # Vibrational constants
            we = c["we"]  # mandatory
            wexe = c.get("wexe", 0)  # optional: 0 is default value
            weye = c.get("weye", 0)
            weze = c.get("weze", 0)
            weae = c.get("weae", 0)
            webe = c.get("webe", 0)

            # Rotational constants
            Be = c["Be"]
            De = c.get("De", 0)

            alpha_e = c.get("alpha_e", 0)
            beta_e = c.get("beta_e", 0)
            gamma_e = c.get("gamma_e", 0)
            delta_e = c.get("delta_e", 0)
            pi_e = c.get("pi_e", 0)

            He = c.get("He", 0)
            eta_e = c.get("eta_e", 0)

            # Energies
            #            Te = c['Te']      # electronic energy
            G = Gv(v, we, wexe, weye, weze, weae, webe)  # vibrational energy
            F = Fv(
                v,
                J,
                Be,
                De,
                alpha_e,
                beta_e,
                gamma_e,
                delta_e,
                pi_e=pi_e,
                He=He,
                eta_e=eta_e,
            )  # rotational energy

            if remove_ZPE:
                ZPE = Gv(0, we, wexe, weye, weze)
            else:
                ZPE = 0

            E = G + F - ZPE  # cm-1

        except KeyError as err:
            raise KeyError(
                "Mandatory spectroscopic constant `{0}` ".format(err.args[0])
                + "not defined for electronic state {0}".format(self.get_fullname())
                + ". Check your ElectronicState definition"
            )

        return E

    # %% Informative methods

    def get_Morse_inc(self):
        r"""Get Morse potential energy increment correction for given molecule

        Examples
        --------

        ::

            inc = ElecState.get_Morse_inc()
            Delta_E(vi, vi+1) = Delta_E0 - (vi+1-vi0)*inc

        for the energy gap between vibrational level vi and vi+1,
        where vi0 is the last vibrational level calculated with Dunham
        expansion, and  and Delta_E0 = DeltaE(vi0-1, vi0)

        See Also
        --------

        :py:func:`~radis.phys.morse.morse_increment`

        """

        from radis.phys.morse import morse_increment

        c = self.rovib_constants

        if get_convention(c) == "dunham":

            # convert to get wexe, weae etc in Herzberg convention
            from radis.db.conventions import herzberg2dunham

            def convert(coef, default=0):
                if coef in c:
                    sign, Yij = herzberg2dunham[coef]
                    return sign * Yij
                else:
                    return default  # default

            we = convert("we")  # required
            wexe = convert("wexe", 0)  # optional: 0 is default value
            weye = convert("weye", 0)
            weze = convert("weze", 0)
            weae = convert("weae", 0)
            webe = convert("webe", 0)
            wece = convert("wece", 0)

            if we is None:
                raise KeyError(
                    "Mandatory spectroscopic constant `we` "
                    + "not defined for electronic state {0}".format(self.get_fullname())
                    + ". Check your ElectronicState definition. Given "
                    + "constants: {0}".format(list(c.keys()))
                )

        else:
            try:
                we = c["we"]  # required
                wexe = c.get("wexe", 0)  # optional: 0 is default value
                weye = c.get("weye", 0)
                weze = c.get("weze", 0)
                weae = c.get("weae", 0)
                webe = c.get("webe", 0)
                wece = c.get("wece", 0)

            except KeyError as err:
                raise KeyError(
                    "Mandatory spectroscopic constant `{0}` ".format(err.args[0])
                    + "not defined for electronic state {0}".format(self.get_fullname())
                    + ". Check your ElectronicState definition. Given "
                    + "constants: {0}".format(list(c.keys()))
                )

        return morse_increment(
            self.Ediss,
            we,
            wexe=wexe,
            weye=weye,
            weze=weze,
            weae=weae,
            webe=webe,
            wece=wece,
        )

    def get_fullname(self):
        return r"{0}({1}{2})(iso{3})".format(
            self.name, self.state, _print_term_symbol(self.term_symbol), self.iso
        )

    def get_statename_utf(self):
        r"""Returns Electronic state name in UTF format, used in the standard
        RADIS JSON databases. Examples::

           'X', '1Σ+'
           >>> X1SIG+

           'C', '3Πu'
           >>> C3PIu

        """

        return self.state + _format_term_symbol(self.term_symbol)


# %% Test


if __name__ == "__main__":

    from radis.test.db.test_molecules import _run_testcases

    print(("Testing molecules.py", _run_testcases()))



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 14 21:10:31 2017

@author: erwan

Plot Absorbance (cm-1) of CO with our code, database HITRAN 2012 against Specair,
under non-equilibrium conditions
"""

from os.path import join

import matplotlib.pyplot as plt
import numpy as np
import pytest

from radis import SpectrumFactory, load_spec, plot_diff
from radis.misc.printer import printm
from radis.test.utils import getValidationCase, setup_test_line_databases


@pytest.mark.fast
def test_validation_vs_specair(rtol=1e-2, verbose=True, plot=False, *args, **kwargs):
    """Test RADIS output on CO IR bands against SPECAIR

    Test is only performed on integrals of absorption coefficient

    RADIS doesnt actually match Specair exactly, but this is due to line intensity
    differences (Specair has no rovibrational specific intensities) rather than
    differences in populations calculations, as evidenced by the partition functions
    comparison in the RADIS presentation article.

    """

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    # %% Specair calculation
    # -----------

    specair_300_300 = load_spec(
        getValidationCase(
            join(
                "test_validation_vs_specair_noneqCO_data",
                "specair_CO_IR_Tvib300_Trot300.spec",
            )
        ),
        binary=True,
    )
    specair_300_2000 = load_spec(
        getValidationCase(
            join(
                "test_validation_vs_specair_noneqCO_data",
                "specair_CO_IR_Tvib300_Trot2000.spec",
            )
        ),
        binary=True,
    )
    specair_2000_300 = load_spec(
        getValidationCase(
            join(
                "test_validation_vs_specair_noneqCO_data",
                "specair_CO_IR_Tvib2000_Trot300.spec",
            )
        ),
        binary=True,
    )

    article_version = False  # just for the article

    # %% Compare with RADIS
    # ----------

    wstep = 0.002
    pl = SpectrumFactory(
        wavelength_min=4400,
        wavelength_max=4900,
        mole_fraction=1,
        pressure=0.01,  # bar
        path_length=1,  # we dont care for abscoeff anyway
        cutoff=1e-30,
        wstep=wstep,
        isotope=1,  # '1,2,3',
        medium="vacuum",
        export_lines=True,
    )  # 0.2)
    pl.warnings["MissingSelfBroadeningWarning"] = "ignore"

    if article_version:
        pl.load_databank("HITEMP-CO-DUNHAM")
    else:
        pl.load_databank("HITRAN-CO-TEST")
        # Available on all systems, convenient for fast testing, but you
        # will be missing some small lines for T ~ 2000 K .
        print(
            "Using HITRAN: small lines will be missing at T ~ 2000K. Use HITEMP if you want them"
        )

    s_300_300 = pl.eq_spectrum(300, name="RADIS")

    s_2000_300 = pl.non_eq_spectrum(Tvib=2000, Trot=300, Ttrans=300, name="RADIS")

    s_300_2000 = pl.non_eq_spectrum(Tvib=300, Trot=2000, Ttrans=2000, name="RADIS")

    # %% Test

    # Compare integrals

    b1 = np.isclose(
        specair_300_300.get_integral("abscoeff"),
        s_300_300.get_integral("abscoeff"),
        rtol=rtol,
    )
    b1 *= np.isclose(
        specair_2000_300.get_integral("abscoeff"),
        s_2000_300.get_integral("abscoeff"),
        rtol=rtol,
    )
    b1 *= np.isclose(
        specair_300_2000.get_integral("abscoeff"),
        s_300_2000.get_integral("abscoeff"),
        rtol=rtol,
    )

    # Compare partition functions to hardcoded values
    b2 = np.isclose(s_2000_300.lines.attrs["Q"], 139, atol=1)  # Specair: 139
    b2 *= np.isclose(s_300_2000.lines.attrs["Q"], 727, atol=1)  # Specair: 727

    if verbose:
        printm(
            ">>> comparing RADIS vs SPECAIR on CO: integrals of abscoeff is are close"
            + " to within {0:.1f}%: {1} ({2:.1f}%, {3:.1f}%, {4:.1f}%)".format(
                rtol * 100,
                bool(b1),
                abs(
                    specair_300_300.get_integral("abscoeff")
                    / s_300_300.get_integral("abscoeff")
                    - 1
                )
                * 100,
                abs(
                    specair_2000_300.get_integral("abscoeff")
                    / s_2000_300.get_integral("abscoeff")
                    - 1
                )
                * 100,
                abs(
                    specair_300_2000.get_integral("abscoeff")
                    / s_300_2000.get_integral("abscoeff")
                    - 1
                )
                * 100,
            )
        )
        printm(
            ">>> comparing RADIS vs SPECAIR on CO: partition functions "
            + "are equal to round error: {0}".format(bool(b2))
        )

    if plot:
        plot_diff(
            specair_300_300,
            s_300_300,
            title=r"T$_\mathregular{vib}$ 300 K, T$_\mathregular{rot}$ 300 K",
            diff_window=int(
                0.02 // wstep
            ),  # compensate for small shifts in both codes. we're comparing intensities here.
            lw_multiplier=1,  # 0.75,
            wunit="nm_vac",
            plot_medium=True,
        )
        plt.xlim((4500, 4900))
        if article_version:
            plt.savefig("out/test_validation_vs_specair_noneqCO_Tvib300_Trot300.png")
            plt.savefig("out/test_validation_vs_specair_noneqCO_Tvib300_Trot300.pdf")

        plot_diff(
            specair_2000_300,
            s_2000_300,
            title=r"T$_\mathregular{vib}$ 2000 K, T$_\mathregular{rot}$ 300 K",
            diff_window=int(
                0.02 // wstep
            ),  # compensate for small shifts in both codes. we're comparing intensities here.
            lw_multiplier=1,  # 0.75,
            wunit="nm_vac",
            plot_medium=True,
        )
        plt.xlim((4500, 4900))
        if article_version:
            plt.savefig("out/test_validation_vs_specair_noneqCO_Tvib2000_Trot300.png")
            plt.savefig("out/test_validation_vs_specair_noneqCO_Tvib2000_Trot300.pdf")

        plot_diff(
            specair_300_2000,
            s_300_2000,
            title=r"T$_\mathregular{vib}$ 300 K, T$_\mathregular{rot}$ 2000 K",
            diff_window=int(
                0.02 // wstep
            ),  # compensate for small shifts in both codes. we're comparing intensities here.
            lw_multiplier=1,  # 0.75,
            wunit="nm_vac",
            plot_medium=True,
        )
        plt.xlim((4500, 4900))
        if article_version:
            plt.savefig("out/test_validation_vs_specair_noneqCO_Tvib300_Trot2000.png")
            plt.savefig("out/test_validation_vs_specair_noneqCO_Tvib300_Trot2000.pdf")

    return bool(b1 * b2)


if __name__ == "__main__":
    printm(
        "test_validation_vs_specair_noneqCO:",
        test_validation_vs_specair(3e-2, plot=True, verbose=True),
    )



# -*- coding: utf-8 -*-
"""Created on Tue Jul 18 17:47:48 2017.

@author: erwan

Dunham development for diatomic molecules energies

Warning
-------

Although by convention prefactors share the same name throughout most of the literature,
signs can be different depending on the article. Make sure your Dunham expansion
has the same signs as the one we use here!

Reference
------------
"Optical Diagnostics and Radiative Emission of Air Plasmas", C. Laux, 1993, p93
Mantz et al 1975, "Ground state molecular constants of 12C16O"
"""


# %% Dunham development

# ... vibrational term (Herzberg notation)


def Gv(v, we=0, wexe=0, weye=0, weze=0, weae=0, webe=0, gv=1):
    """Vibrational energy term Dunham development (order 5 in v) in Herzberg
    notation.

    .. math::

        G_v = w_e\\left(v+\\frac{g_v}{2}\\right) - w_ex_e\\left(v+\\frac{g_v}{2}\\right)^2 +
        w_ey_e\\left(v+\\frac{g_v}{2}\\right)^3  + w_ez_e\\left(v+\\frac{g_v}{2}\\right)^4 +
        w_ea_e\\left(v+\\frac{g_v}{2}\\right)^5

    Parameters
    ----------

    v: int
        vibrational quantum number
    we: float
            vibrational constant – first term (cm-1)
    ωexe: float
            vibrational constant – second term (cm-1)
    ωeye: float
            vibrational constant – third term (cm-1)
    ωeze: float
            vibrational constant – fourth term (cm-1)
    weae: float
       vibrational constant – fifth term (cm-1)
    webe:
        vibrational constant – sixth term (cm-1)
    gv: int
       degeneracy   (usually 1, but 2 for CO2-v2)

    Returns
    -------

    Gv: float
        Energy (cm-1)


    Notes
    -----

    Validity:

    For large vibrational levels Dunham's expansion is not valid. In other LBL codes,
    such as Specair, Morse Potential is used above a certain vibrational level.
    See :func:`~radis.phys.morse.morse_increment` for an implementation in RADIS.

    .. warning::

        Although by convention prefactors share the same name throughout most of the literature,
        signs can be different depending on the article. Make sure your Dunham expansion
        has the same signs as the one we use here! In particular, this function
        uses a ``-wexe`` sign as in the usual Herzberg notation. Ex:

        - Mantz and Maillard 1975 (CO X)  uses opposite signs for **weze** and **webe**

    References
    ----------

    "Optical Diagnostics and Radiative Emission of Air Plasmas", C. Laux, 1993, p93

    See Also
    --------

    :py:func:`~radis.levels.dunham.Fv`

    List of :py:data:`~radis.db.conventions.herzberg_coefficients`

    Get Herzberg coefficients for a molecule: :py:func:`~radis.db.utils.get_herzberg_coefficients`

    Use Dunham coefficients instead: :py:func:`~radis.levels.dunham.EvJ`
    """

    return (
        we * (v + gv / 2)
        - wexe * (v + gv / 2) ** 2
        + weye * (v + gv / 2) ** 3
        + weze * (v + gv / 2) ** 4
        + weae * (v + gv / 2) ** 5
        + webe * (v + gv / 2) ** 6
    )  # + ..
    #                :                                :
    #                :                                :
    # In Mantz and Maillard 1975                      - weze*(v+gv/2)**4
    # Mantz 1975     - webe*(v+gv/2)**6


# ... rotational term (Herzberg notation)


def Fv(
    v,
    J,
    Be=0,
    De=0,
    alpha_e=0,
    beta_e=0,
    gamma_e=0,
    delta_e=0,
    epsilon_e=0,
    pi_e=0,
    He=0,
    eta_e=0,
    gv=1,
):
    """Rotational energy term Dunham development (order 4 in J) in Herzberg
    notation.

    .. math::
        B_{v}=B_{e}-\\alpha_{e}\\left(v+\\frac{g_{v}}{2}\\right)+\\gamma_{e}
        \\left(v+\\frac{g_{v}}{2}\\right)^{2}+\\delta_{e}\\left(v+\\frac{g_{v}}{2}
        \\right)^{3}+\\epsilon_{e}\\left(v+\\frac{g_{v}}{2}\\right)^{4}

        D_{v}=D_{e}+\\beta_{e}\\left(v+\\frac{g_{v}}{2}\\right)+\\pi_{e}
        \\left(v+\\frac{g_{v}}{2}\\right)^{2}

        H_{v}=H_{e}-\\eta_{e}\\left(v+\\frac{g_{v}}{2}\\right)

    *generated from the Python formula with* :py:func:`~pytexit.pytexit.py2tex`


    Parameters
    ----------

    v: int
        vibrational quantum number
    J: int
        rotational quantum number
    Be: float
            rotational constant in equilibrium position (cm-1)
    De: float
            centrifugal distortion constant (cm-1)
    alpha_e: float
            rotational constant – first term (cm-1)
    beta_e: float
            rotational constant – first term, centrifugal force (cm-1)
    gamma_e: float
            rotation-vibration interaction constant (cm-1)
    delta_e: float
            (cm-1)
    epsilon_e: float
       (cm-1)
    pi_e: float
        (cm-1)
    He: float
        third order correction factor (cm-1)
    eta_e: float
        (float)
    gv: int
       degeneracy   (usually 1, but 2 for CO2-v2)

    Returns
    -------

    Fv: float
        Energy (cm-1)

    Notes
    -----

    Validity:

    For large vibrational levels Dunham's expansion is not valid. In RADIS a
    Morse Potential can be used above a certain vibrational level

    .. warning::

        Although by convention prefactors share the same name throughout most of the literature,
        signs can be different depending on the article. Make sure your Dunham expansion
        has the same signs as the one we use here! Ex:

        - Mantz and Maillard 1975  (CO)  uses opposite signs for **delta_e** and **beta_e**

    References
    ----------

    "Optical Diagnostics and Radiative Emission of Air Plasmas", C. Laux, 1993, p93

    See Also
    --------

    :py:func:`~radis.levels.dunham.Gv`

    List of :py:data:`~radis.db.conventions.herzberg_coefficients`

    Get Herzberg coefficients for a molecule: :py:func:`~radis.db.utils.get_herzberg_coefficients`

    Use Dunham coefficients instead: :py:func:`~radis.levels.dunham.EvJ`
    """

    B_e, D_e, H_e, g_v = Be, De, He, gv

    # ... Note: formula added in docstring with pytexit.py2tex()
    # + ...
    B_v = (
        B_e
        - alpha_e * (v + g_v / 2)
        + gamma_e * (v + g_v / 2) ** 2
        + delta_e * (v + g_v / 2) ** 3
        + epsilon_e * (v + g_v / 2) ** 4
    )
    D_v = D_e + beta_e * (v + g_v / 2) + pi_e * (v + g_v / 2) ** 2  # + ...
    H_v = H_e - eta_e * (v + g_v / 2)
    #       :                                        :
    #       :                                        :
    # In Mantz and Maillard 1975                     - delta_e*(v+gv/2)**3
    #       - beta_e*(v+gv/2)      in Mantz and Maillard 1975

    return B_v * J * (J + 1) - D_v * (J * (J + 1)) ** 2 + H_v * (J * (J + 1)) ** 3


# ... general term

# def EvJ(v, J, **Ykl_dict):
#    ''' Calculates rovibrational energy reading from Dunham coefficients in
#    Ykl notation
#
#    Parameters
#    ----------
#
#    Ykl: dict
#        an arbitrary dictionary of Ykl coefficients
#        accepted formats: Y01, Y01_cm-1
#
#    Ykl are parsed and assigned the correct energy
#
#    Notes
#    -----
#
#    Because reading and parsing the Ykl dict is required, this is expected to
#    be slower that a hardcoded implementation. However, it is also much
#    more flexible.
#
#
#    Examples
#    --------
#
#    Read directly from a .json file::
#
#        from radis.db.utils import get_dunham_coefficients
#        from radis.levels.dunham import EvJ
#        dunham_coeffs = get_dunham_coefficients('CO', 1, 'X1SIG+')
#
#        # Now calculate energy
#        EvJ(v=0, J=0, **dunham_coeffs)
#
#    '''
#
#    E = 0
#
#    Ykl_format = '^Y(?P<k>[0-9])(?P<l>[0-9])(_cm-1)?$'
#    # ... Ykl with k, l ints  and followed by nothing or _cm-1
#    # ... accepted formats: Y01, Y01_cm-1
#    reg = re.compile(Ykl_format)
#
#    for ykl_name, Ykl in Ykl_dict.items():
#        match = reg.search(ykl_name)
#        if match is None:
#            raise ValueError('Key {0} does not have the expected format {1}'.format(
#                    ykl_name, Ykl_format))
#        res = match.groupdict()
#        k = int(res['k'])
#        l = int(res['l'])
#        E += Ykl * (v+0.5)**k * (J*(J+1))**l
#
#    return E

# New version (less versatile, but less Python and faster)
# only allowed formats: Y01
def EvJ(v, J, **Ykl_dict):
    """Calculates rovibrational energy reading from Dunham coefficients in Ykl
    notation, for diatomic molecules.

    Parameters
    ----------

    Ykl: dict
        an arbitrary dictionary of Ykl coefficients
        accepted formats: Y01

    Ykl are parsed and assigned the correct energy

    Examples
    --------

    Read directly from a .json file::

        from radis.db.utils import get_dunham_coefficients
        from radis.levels.dunham import EvJ
        dunham_coeffs = get_dunham_coefficients('CO', 1, 'X1SIG+')

        # Now calculate energy
        EvJ(v=0, J=0, **dunham_coeffs)

    See Also
    --------

    Get Dunham coefficients for a molecule: :py:func:`~radis.db.utils.get_dunham_coefficients`

    Use Herzberg coefficients instead: :py:func:`~radis.levels.dunham.Gv`,
    :py:func:`~radis.levels.dunham.Fv`
    """

    E = 0

    for ykl_name, Ykl in Ykl_dict.items():
        k = int(ykl_name[1])
        l = int(ykl_name[2])
        E += Ykl * (v + 0.5) ** k * (J * (J + 1)) ** l

    return E


if __name__ == "__main__":

    from radis.test.phys.test_dunham import _run_all_tests

    print("Testing Dunham.py: ", _run_all_tests())



# -*- coding: utf-8 -*-
"""
Summary
-----------------------

Kurucz database parser

Based largely on the `Exojax <https://github.com/HajimeKawahara/exojax>`__ code

-----------------------


"""

import io
import os
from os.path import abspath, expanduser, join, splitext

import numpy as np
import pandas as pd
import periodictable

from radis.api.dbmanager import DatabaseManager
from radis.db.classes import get_ielem_charge, roman_to_int
from radis.misc.config import getDatabankEntries
from radis.misc.utils import getProjectRoot
from radis.misc.warning import DatabaseAlreadyExists
from radis.phys.air import air2vacuum
from radis.phys.constants import c_CGS, e_CGS, m_e_CGS


def pick_ionE(ielem, charge):
    # This method was extracted from exojax/src/exojax/spec/atomllapi.py
    # (https://github.com/HajimeKawahara/exojax.git)

    """Pick out the ionization energy of a specific atomic species from NIST.

    Parameters
    ----------
    ielem: int
        atomic number (e.g., Fe=26)
    charge: int
        charge of the species

    Returns
    -------
    ionE: float
        ionization energy

    """

    fn_IonE = os.path.join(
        getProjectRoot(), "db", "NIST_Atomic_Ionization_Energies.txt"
    )  # pkgutil.get_data(
    #     "exojax", "data/atom/NIST_Atomic_Ionization_Energies.txt"
    # )
    df_ionE = pd.read_csv(fn_IonE, sep="|", skiprows=6, header=0)

    def f_droppare(x):
        return (
            x.str.replace("\(", "", regex=True)
            .str.replace("\)", "", regex=True)
            .str.replace("\[", "", regex=True)
            .str.replace("\]", "", regex=True)
            .str.replace("                                      ", "0", regex=True)
        )

    ionE = f_droppare(
        df_ionE[(df_ionE["At. num "] == ielem) & (df_ionE[" Ion Charge "] == charge)][
            "      Ionization Energy (a) (eV)      "
        ]
    )

    assert ionE.size == 1

    return float(ionE.iloc[0])


def get_atomic_number(species):
    """
    Extracts the atomic number from the species given in spectroscopic notation
    """
    atomic_symbol = species.split("_")[0]
    el = getattr(periodictable, atomic_symbol)

    atomic_number = el.number
    return atomic_number


def get_ionization_state(species):
    """
    Extracts the ionization state from the species given in spectroscopic notation
    """
    ionization_str = species.split("_")[1]
    ionization_int = roman_to_int(ionization_str) - 1
    # formatted_str = f"{ionization_int:02}"

    return ionization_int  # formatted_str


def get_element_symbol(species):
    """
    Extracts the element symbol from the species given in spectroscopic notation
    """

    atomic_symbol = species.split("_")[0]
    el = getattr(periodictable, atomic_symbol)
    return el


def read_kurucz_pfs(file):
    """Convert a Kurucz partfnxxyy.dat file containing a table of partition functions by temperature and potential lowering to a Pandas DataFrame

    Parameters
    ----------
    fname: str
        file name

    Returns
    ----------
    df: Pandas DataFrame
    """
    df = pd.read_csv(file, sep="\s+", header=2)
    df.set_index("LOG10(T)", inplace=True)
    return df


def load_pf_Barklem2016():
    """Load a table of the partition functions for 284 atomic species.

    Returns
    -------
    pfTdat:  pd.DataFrame
        Steps of temperature (K)
    pfdat:  pd.DataFrame
        Partition functions for 284 atomic species

    References
    ----------
    `Barklem & Collet (2016), Table 8 <https://doi.org/10.1051/0004-6361/201526961>`_

    """
    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.join(current_dir, "./../db/pfTKurucz_values.txt")
    with open(file_path, "r") as file:
        pfT_str = file.read()
    pfTdat = pd.read_csv(io.StringIO(pfT_str), sep="\s+")
    pfTdat = pd.Series(pfTdat.columns[1:]).astype(
        "float64"
    )  # Converts the values to float64, skipping the first value

    with open(os.path.join(getProjectRoot(), "db", "kuruczpartfn.txt"), "r") as f:
        pfdat = pd.read_csv(f, sep="\s+", comment="#", names=pfTdat.index)

    return pfTdat, pfdat


class KuruczDatabaseManager(DatabaseManager):
    def __init__(
        self,
        name,
        molecule,
        local_databases,
        engine="default",
        verbose=True,
        parallel=True,
    ):
        super().__init__(
            name,
            molecule,
            local_databases,
            engine,
            verbose=verbose,
            parallel=parallel,
        )
        self.wmin = None
        self.wmax = None

        self.atomic_number = f"{get_atomic_number(molecule):02}"
        self.ionization_state = f"{get_ionization_state(molecule):02}"

        self.actual_file = None
        self.actual_url = None
        self.pf_path = None

    def fetch_urlnames(self):
        """returns the possible urls of the desired linelist file, pending confirmation upon attempting download as to which is correct"""

        code = f"{self.atomic_number}{self.ionization_state}"

        urlnames = [
            "http://kurucz.harvard.edu/atoms/"
            + code
            + "/gf"
            + code
            + ".all",  # 1. new linelist including lab lines
            "http://kurucz.harvard.edu/atoms/"
            + code
            + "/gf"
            + code
            + ".pos",  # 2. new linelist without lab lines
            "http://kurucz.harvard.edu/linelists/gfall/gf"
            + code
            + ".all",  # 3. old linelist including lab lines
        ]

        return urlnames

    def get_pf_path(self):
        """returns the url at which the dedicated partition function file is expected to be located, and the derived file path at which it would be saved after downloading and parsing"""
        code = f"{self.atomic_number}{self.ionization_state}"
        url = "http://kurucz.harvard.edu/atoms/" + code + "/partfn" + code + ".dat"
        fname = splitext(url.split("/")[-1])[0] + ".h5"
        path = expanduser(
            abspath(join(self.local_databases, self.molecule + "-" + fname))
        )
        return path, url

    def get_possible_files(self, urlnames=None):
        """returns the urls from fretch_urlnames and the derived file paths at which each would be saved after downloading and parsing"""
        verbose = self.verbose
        local_databases = self.local_databases
        engine = self.engine

        # copied from DatabaseManager.get_filenames:
        if urlnames is None:
            urlnames = self.fetch_urlnames()
        local_fnames = [
            (
                splitext(splitext(url.split("/")[-1])[0])[0]  # twice to remove .par.bz2
                + ".h5"
            )
            for url in urlnames
        ]

        try:
            os.mkdir(local_databases)
        except OSError:
            pass
        else:
            if verbose:
                print("Created folder :", local_databases)

        local_files = [
            abspath(
                join(
                    local_databases,
                    self.molecule + "-" + local_fname,
                )
            )
            for local_fname in local_fnames
        ]

        if engine == "vaex":
            local_files = [fname.replace(".h5", ".hdf5") for fname in local_files]

        local_files = [expanduser(f) for f in local_files]

        return local_files, urlnames

    def parse_to_local_file(
        self,
        opener,
        urlname,
        local_file,
        pbar_active=True,
        pbar_t0=0,
        pbar_Ntot_estimate_factor=None,
        pbar_Nlines_already=0,
        pbar_last=True,
    ):
        """overwrites parent class method as required"""

        writer = self.get_datafile_manager()
        if local_file == self.pf_path:
            df = read_kurucz_pfs(opener.abspath(urlname))
            with pd.HDFStore(local_file, mode="w", complib="blosc", complevel=9) as f:
                f.put(key="df", value=df)
            Nlines = len(df)  # irrelevant
        else:
            df = read_kurucz(opener.abspath(urlname))

            writer.write(local_file, df, append=False)

            self.wmin = df.wav.min()
            self.wmax = df.wav.max()

            Nlines = len(df)

            writer.combine_temp_batch_files(local_file)

            # Add metadata
            from radis import __version__

            writer.add_metadata(
                local_file,
                {
                    "wavenumber_min": self.wmin,
                    "wavenumber_max": self.wmax,
                    "download_date": self.get_today(),
                    "download_url": urlname,
                    "total_lines": Nlines,
                    "version": __version__,
                },
            )

        return Nlines

    def register(self, get_main_files, get_pf_files):

        if self.is_registered():
            dict_entries = getDatabankEntries(
                self.name
            )  # just update previous register details
        else:
            dict_entries = {}

        if get_main_files:
            files = [self.actual_file]
            urls = [self.actual_url]
            dict_entries.update(
                {
                    "path": files,
                    "format": "hdf5-radisdb",
                    "download_date": self.get_today(),
                    "download_url": urls,
                }
            )
            if self.wmin and self.wmin:
                info = f"Kurucz {self.molecule} lines ({self.wmin:.1f}-{self.wmax:.1f} cm-1)."
                dict_entries.update(
                    {
                        "info": info,
                        "wavenumber_min": self.wmin,
                        "wavenumber_max": self.wmax,
                    }
                )
        # else:
        #     raise Exception("Kurucz database can't be registered until the correct url to use is known")

        if get_pf_files:
            # files.append(self.pf_path)
            # urls.append(self.pf_url)
            dict_entries.update(
                {
                    "parfunc": self.pf_path,
                    "parfuncfmt": "kurucz",
                }
            )

        dict_entries.update(
            {
                "parfuncfmt": "kurucz",
            }
        )

        try:
            super().register(dict_entries)
        except DatabaseAlreadyExists as e:
            raise Exception(
                'If you want RADIS to overwrite the existing entry for a registered databank, set the config option "ALLOW_OVERWRITE" to True.'
            ) from e


def read_kurucz(kuruczf, preserve_orig_levels=False):
    """
    Parse a Kurucz linelist, process its columns as required, and return as a Pandas DataFrame

    Inspired by: https://github.com/rasmus98/NLTE-Helium/blob/fc6161a30ebecfcf59f36386b1dc7f02ff749905/Grotrian_diagram_Helium.ipynb cell 2. Also see:

    - https://github.com/DBerke/varconlib/blob/e57250ca359026ae8b8059dae179fb0ad9625aa2/varconlib/scripts/select_line_pairs.py#L711

    - https://github.com/followthesheep/specutils/blob/e2873719f3820e83ab29c79062d7a59a7664fa2f/specutils/read_kurucz_linelist.py

    - https://github.com/ajwheeler/Korg.jl/blob/53420d38c23e21a7fe202b8680bdb201c9a62a2a/src/linelist.jl#L153

    Parameters
    ----------
    kuruczf: str
        file path
    preserve_orig_levels: boolean
        whether to preserve the original columns pertaining to the levels prior to transforming them, with '_orig' appended to their name; note the columns are still reversed so that the database index increases in wavenumber rather than wavelength

    Returns
    ----------
        Pandas DataFrame containing the required columns for the Kurucz linelist database
    """
    colspecs = (
        (0, 11),
        (11, 18),
        (18, 24),
        (24, 36),
        (36, 41),
        (42, 52),
        (52, 64),
        (64, 69),
        (70, 80),
        (80, 86),
        (86, 92),
        (92, 98),
        # (98,102),(102,104),(104,106),
        (106, 109),
    )
    # (109,115),(115,118),(118,124),(124,129),(129,134),(135,136),(136,137),(138,139),(139,140),(140,141),(141,144),(144,149),(149,154),(154,160))
    dtypes = (
        np.float64,
        np.float64,
        "object",  # make species object rather than float for easier parsing as a string
        np.float64,
        np.float64,
        "object",
        np.float64,
        np.float64,
        "object",
        np.float64,
        np.float64,
        np.float64,
        # str,np.int64,np.int64,
        np.int64,
    )
    # np.float64,np.int64,np.float64,np.int64,np.int64,str,str,str,str,np.int64,str,np.int64,np.int64,np.int64)
    colnames = (
        "orig_wavelen",
        "loggf",
        "species",
        "elower_orig",
        "jlower_orig",
        "labellower_orig",
        "eupper_orig",
        "jupper_orig",
        "labelupper_orig",
        "gamRad",
        "gamSta",
        "gamvdW",
        #'ref','NLTElo','NLTEhi',
        "iso",
    )
    #'hyperstrength','isotope2','isoabund','hyperlo','hyperhi','hyperFlo','hypernotelo','hyperFhi','hypernotehi','strengthclassnote','trans_code','lande_g_even','lande_g_odd','iso_shift')
    df = pd.read_fwf(
        kuruczf,
        colspecs=colspecs,
        dtype=dict(zip(colnames, dtypes)),
        names=colnames,
        skip_blank_lines=True,
    )  # , converters={'labellower_orig': str.strip, 'labelupper_orig': str.strip}) # appears to strip automatically
    species_unique = df["species"].unique()
    assert len(species_unique) == 1
    ielem, charge = get_ielem_charge(species_unique[0])
    df["ionE"] = pick_ionE(ielem, charge)

    # simple method to preserve dtypes of each column:
    cond = (df["eupper_orig"] - df["elower_orig"]) > 0
    oldlower = ["jlower_orig", "labellower_orig", "elower_orig"]
    oldupper = ["jupper_orig", "labelupper_orig", "eupper_orig"]
    newlower = ["jl", "labellower", "El"]
    newupper = ["ju", "labelupper", "Eu"]
    for oldcol1, oldcol2, newcol1, newcol2 in zip(
        oldlower, oldupper, newlower, newupper
    ):
        df[newcol1] = df[oldcol1].where(cond, df[oldcol2])
        df[newcol2] = df[oldcol2].where(cond, df[oldcol1])
    # df1[['jlower_orig', 'jupper_orig', 'labellower_orig', 'labelupper_orig', 'elower_orig', 'eupper_orig']] = df[['jupper_orig', 'jlower_orig', 'labelupper_orig', 'labellower_orig', 'eupper_orig', 'elower_orig']]
    # df.where(cond, df1, inplace=True) # returns TypeError: Cannot do inplace boolean setting on mixed-types with a non np.nan value
    # np.where(pd.concat([cond]*6, axis=1), df[['jlower_orig', 'jupper_orig', 'labellower_orig', 'labelupper_orig', 'elower_orig', 'eupper_orig']], df[['jupper_orig', 'jlower_orig', 'labelupper_orig', 'labellower_orig', 'eupper_orig', 'elower_orig']]) # doesn't preserve dtype

    df["wav"] = 1e7 / df["orig_wavelen"].where(
        df["orig_wavelen"] < 200, air2vacuum(df["orig_wavelen"])
    )
    df["gu"] = df["ju"] * 2 + 1
    df["A"] = (
        10 ** df["loggf"]
        / df["gu"]
        * (c_CGS * df["wav"]) ** 2
        * (8 * np.pi**2 * e_CGS**2)
        / (m_e_CGS * c_CGS**3)
    )

    df[:] = df[::-1]

    if not preserve_orig_levels:
        df.drop(
            [
                "elower_orig",
                "eupper_orig",
                "jlower_orig",
                "jupper_orig",
                "labellower_orig",
                "labelupper_orig",
            ],
            axis=1,
            inplace=True,
        )

    return df



# -*- coding: utf-8 -*-
"""
Created on Sat May  1 13:50:36 2021

@author: erwan
"""

import numpy as np

import radis

KNOWN_CONTEXT = ["paper", "notebook", "talk", "poster"]


def _get_defaults(plotlib, context, style):

    expected_format = {
        "plot": {
            "plotlib": "ANY OF "
            + "/".join(['"publib"', '"seaborn"', '"matplotlib"', '"none"']),
            "context": "ANY OF " + "/".join(f'"{c}"' for c in KNOWN_CONTEXT),
            "style": "ANY OF the publib / seaborn / matplotlib themes",
        }
    }

    # Get defaults
    if plotlib == "config-default":
        try:
            plotlib = radis.config["plot"]["plotlib"]
        except KeyError:
            raise ValueError(
                "Missing key in your ~/radis.json user config file. Expected format :\n\n {0}".format(
                    expected_format
                )
            )

    if context == "config-default":
        try:
            context = radis.config["plot"]["context"]
        except KeyError:
            raise ValueError(
                "Missing key `context` in your ~/radis.json user config file. Expected format :\n\n {0}".format(
                    expected_format
                )
            )

    if style == "config-default":
        try:
            style = radis.config["plot"]["style"]
        except KeyError:
            raise ValueError(
                "Missing key `style` in your ~/radis.json user config file. Expected format :\n\n {0}".format(
                    expected_format
                )
            )

    if context and context not in KNOWN_CONTEXT:
        raise ValueError(f"context must be in {KNOWN_CONTEXT}. Got {context}")

    if plotlib == "publib":
        if not isinstance(style, list):
            style = [style]
        if context == "paper":
            style = style + ["article"]
        elif context in ["poster", "notebook", "talk"]:
            style = style + [context]

    return plotlib, context, style


def set_style(
    plotlib="config-default", context="config-default", style="config-default"
):
    """Set styles used by Radis plot functions (:py:func:`~radis.spectrum.spectrum.Spectrum.plot`,
    :py:func:`~radis.spectrum.compare.plot_diff`, :py:func:`~radis.tools.slit.plot_slit`, etc.)

    Parameters
    ----------
    plotlib: ``"publib"``, ``"seaborn"``, ``"matplotlib"``
        `publib <https://github.com/erwanp/publib>`__ , :py:mod:`seaborn`, :py:mod:`matplotlib`
    context: ``"poster"``, ``"talk"``, ``"paper"``, ``"notebook"``
        See :py:func:`seaborn.set_theme`
    style: a ``publib`` or ``matplotlib`` style, or a ``seaborn`` theme
        if ``"oring"``, use your current defaults.
        if ``"none"``, use your current defaults.

    Examples
    --------

    .. minigallery:: radis.misc.plot.set_style

    See also
    --------
    :py:func:`seaborn.set_theme`,
    :py:func:`publib.main.set_style`
    """

    plotlib, context, style = _get_defaults(plotlib, context, style)

    if plotlib != "none":
        import matplotlib

        matplotlib.rc_file_defaults()

    # Apply style
    if plotlib == "publib":
        import publib

        publib.set_style(style)
    elif plotlib == "seaborn":
        import seaborn as sns

        sns.set_theme(context=context, style=style)
    elif plotlib == "matplotlib":
        import matplotlib.pyplot as plt

        plt.style.use(style)
    elif plotlib == "none":
        pass
    else:
        raise ValueError(f"plot library: {plotlib}")


def fix_style(
    plotlib="config-default",
    context="config-default",
    style="config-default",
    *args,
    **kwargs,
):
    """A posteriori improvement of the last figure generated

    Effective when ``plotlib`` is `publib <https://github.com/erwanp/publib>`__

    Parameters
    ----------
    plotlib: ``"publib"``, ``"seaborn"``, ``"matplotlib"``
        `publib <https://github.com/erwanp/publib>`__ , :py:mod:`seaborn`, :py:mod:`matplotlib`
    context: ``"poster"``, ``"talk"``, ``"paper"``, ``"notebook"``
        See :py:func:`seaborn.set_theme`
    style: a ``publib`` or ``matplotlib`` style, or a ``seaborn`` theme
        if ``"oring"``, use your current defaults.
        if ``"none"``, use your current defaults.

    Examples
    --------

    .. minigallery:: radis.misc.plot.fix_style

    See also
    --------
    :py:func:`publib.main.fix_style`
    """

    plotlib, context, style = _get_defaults(plotlib, context, style)

    # Apply style
    if plotlib == "publib":
        import publib

        publib.fix_style(style, *args, **kwargs)
    elif plotlib == "seaborn":
        pass
    elif plotlib == "matplotlib":
        pass
    elif plotlib == "none":
        pass
    else:
        raise ValueError(f"plot library: {plotlib}")


def split_and_plot_by_parts(w, I, *args, **kwargs):
    """Plot two discontinued arrays (typically a spectrum) without showing
    junctions: first identify junctions then split and plot separately.

    Useful for plotting an experimental spectrum defined on different, non overlapping
    ranges without showing connecting lines between the ranges, or to plot an
    experimental spectrum defined on overlapping ranges, without showing connecting
    lines neither.

    Parameters
    ----------
    w, I: arrays
        typically output of :py:func:`~numpy.hstack`.

    Other Parameters
    ----------------
    split_threshold: int
        number of standard deviation for threshold. Default 10
    ax: matplotlib axe
        plot on a particular axe
    kwargs: dict
        forwarded to :func:`~matplotlib.pyplot.plot`
    cutwings: int
        discard elements on the side. Default 0

    Returns
    -------
    ax.plot
    """

    import matplotlib.pyplot as plt
    from publib.tools import keep_color

    # Get defaults
    ax = kwargs.pop("ax", None)
    if ax is None:
        ax = plt.gca()
    split_threshold = kwargs.pop("split_threshold", 10)  # type: int
    cutwings = kwargs.pop("cutwings", 0)  # type: int
    label = kwargs.pop("label", None)  # type: str

    # identify joints
    dw = np.diff(w)
    dwmean = dw.mean()
    joints = np.argwhere((abs(dw - dwmean) > split_threshold * dw.std())) + 1

    # Split
    if len(joints) > 0:
        ws = np.split(w, joints.flatten())
        Is = np.split(I, joints.flatten())

        # Plot separately
        out = []
        for i, (wi, Ii) in enumerate(zip(ws, Is)):
            if cutwings:
                wi = wi[cutwings:-cutwings]
                Ii = Ii[cutwings:-cutwings]
            if i == 0:  # label once only
                out.append(
                    ax.plot(
                        wi, Ii, *args, **dict(list(kwargs.items()) + [("label", label)])
                    )
                )
            else:
                keep_color()
                out.append(ax.plot(wi, Ii, *args, **kwargs))

        return list(zip(*out))

    else:
        if cutwings:
            w = w[cutwings:-cutwings]
            I = I[cutwings:-cutwings]
        return ax.plot(w, I, *args, **dict(list(kwargs.items()) + [("label", label)]))



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 14 15:38:18 2018

@author: erwan
"""


import pytest


@pytest.mark.fast
def test_flatten(*args, **kwargs):
    from radis.misc.basics import flatten

    assert flatten(*[600, 300, (600, 600, 700)]) == [600, 300, 600, 600, 700]


@pytest.mark.fast
def test_round_off(*args, **kwargs):
    from radis.misc.basics import round_off

    assert round_off(0.024231242) == 0.024
    assert round_off(0.0019932133) == 0.002
    assert round_off(0.0000000000002193) == 0


@pytest.mark.fast
def test_str2bool(*args, **kwargs):
    import pandas as pd

    from radis.misc.basics import str2bool

    df = pd.Series(["1.0", "true", "True", 1])
    assert df.dtypes == object

    # Test
    assert df.map(str2bool).all()

    df = pd.Series(["0.0", "false", "False", 0])
    assert df.dtypes == object

    # Test
    assert not df.map(str2bool).any()



# -*- coding: utf-8 -*-
"""
Created on Sun Apr  9 21:37:59 2017

@author: erwan

Reproduce the plasma torch experiment of Packan 2003 in in atmospheric air

Conditions
--------

There is 6 m of room air between the plasma and the detector. The temperature
and humidity in the laboratory were recorded during experiment and calibration.
In both cases the measured ambient temperature was 25°C, and the relative humidity
was 42% (10% uncertainty), which corresponds to a mole fraction of H2O of
1.3+/-0.1 x 1E-2. We used 1.4e-2 because it gave better agreement for absorption
on the fundamental bands of NO. The experiments were done in about 1997, and the
room-air CO2 concentration was assumed to be 330 ppm.

Temperatures and concentrations of all slabs are stored in
`test_compare_torch_CO2_conditions_JTHT2003.dat`



"""

from time import time

import matplotlib.pyplot as plt
import pandas as pd
import pytest

from radis.lbl import SpectrumFactory
from radis.los import MergeSlabs, SerialSlabs
from radis.misc.printer import printm
from radis.phys.convert import nm2cm
from radis.spectrum import experimental_spectrum, get_residual
from radis.test.utils import getValidationCase


@pytest.mark.needs_config_file
@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_db_HITEMP_CO_DUNHAM
def test_compare_torch_CO2(
    verbose=True, plot=False, save=False, warnings=True, use_cache=True, *args, **kwargs
):
    """Reproduce the plasma torch experiment of Packan 2003 in in atmospheric air

    Notes
    -----

    Thresholds parameters are reduced for a faster calculation time. For instance:

    - truncation should be 25 rather than 10
    - cutoff should be 1e-27 rather than 1e-25
    - wstep should be 0.01 or even 0.008 rather than 0.1

    Performance:

    - neq==0.9.20: Finished test_compare_torch_CO2 in 758.640324s

    - neq==0.9.21: Finished test_compare_torch_CO2 in 672s

    - neq==0.9.21*: (with ParallelFactory) Finished test_compare_torch_CO2 in 298s

    - neq==0.9.22: (Parallel + continuum) Finished in 65s
      RADIS 0.9.9 == neq 0.9.24

    - RADIS 0.9.26: Finished test_compare_torch_CO2 in 57s (LDM, no continuum)

    - RADIS 0.9.28: ParallelFactory is removed. Finished test_compare_torch_CO2 in 45s

    Reference
    --------

    Packan et al 2003 "Measurement and Modeling of OH, NO, and CO Infrared Radiation
    at 3400 K", JTHT

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    t0 = time()

    # %% Conditions

    wlmin = 4100
    wlmax = 4900

    wmin = nm2cm(wlmax)
    wmax = nm2cm(wlmin)
    wstep = 0.1

    # Load concentration profile
    conds = pd.read_csv(
        getValidationCase(
            "test_compare_torch_CO2_data/test_compare_torch_CO2_conditions_JTHT2003.dat"
        ),
        comment="#",
        sep="\s+",
    )
    slab_width = 0.05  # cm,  WARNING. Hardcoded (look up the table)

    # %% Calculate slabs

    # CO2
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=None,
        path_length=None,
        molecule="CO2",
        isotope="1,2",
        wstep=wstep,
        export_lines=False,  # saves some memory
        export_populations=False,  # saves some memory
        cutoff=1e-25,
        truncation=10,
        # pseudo_continuum_threshold=0.01, # use pseudo-continuum, no LDM. Note : 56s on 20/08.
        # optimization=None,
        pseudo_continuum_threshold=0,  # use LDM, no pseudo-continuum. Note : 84s on 20/08
        optimization="min-RMS",
        verbose=False,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    #        # Init database: only if we want to save all spectra being calculated
    # (discarded for the moment)
    #        if use_cache:
    #            sf.init_database('test_compare_torch_CO2_SpectrumDatabase_HITEMP_1e25',
    #                             add_info=['Tgas'])
    sf.init_databank(
        "HITEMP-CO2-DUNHAM",
        db_use_cached=use_cache,
    )  # saves some memory
    #    sf.init_databank('CDSD-4000')

    # CO
    sfco = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=None,
        path_length=None,
        molecule="CO",
        isotope="1,2",
        wstep=wstep,
        export_lines=False,  # saves some memory
        export_populations=False,  # saves some memory
        cutoff=1e-25,
        truncation=10,
        optimization=None,
        verbose=False,
    )
    sfco.warnings["MissingSelfBroadeningWarning"] = "ignore"
    # Init database: only if we want to save all spectra being calculated
    # (discarded for the moment)
    #        if use_cache:
    #            sfco.init_database(
    #                'test_compare_torch_CO_SpectrumDatabase_HITEMP_1e25', add_info=['Tgas'])
    sfco.init_databank("HITEMP-CO-DUNHAM", db_use_cached=use_cache)

    # Calculate all slabs
    # .. CO2 at equilibrium
    slabsco2 = []
    for Tgas, xCO2 in zip(conds.T_K, conds.CO2):
        slabsco2.append(
            sf.eq_spectrum(Tgas=Tgas, mole_fraction=xCO2, path_length=slab_width)
        )
    # .. CO at equilibrium
    slabsco = []
    for Tgas, xCO in zip(conds.T_K, conds.CO):
        slabsco.append(
            sfco.eq_spectrum(Tgas=Tgas, mole_fraction=xCO, path_length=slab_width)
        )

    # Room absorption
    s0 = sf.eq_spectrum(Tgas=300, mole_fraction=330e-6, path_length=600)
    # Warning: This slab includes the CO2 emission from the 300 K air
    # within the spectrometer. But experimentally it is subtracted
    # by there the chopper (this is corrected below)
    # ... see RADIS 2019 paper for more details

    # Add radiance for everyone if not calculated (ex: loaded from database)
    for s in slabsco2 + slabsco + [s0]:
        s.update()

    # Merge CO + CO2 for each slab
    slabstot = [MergeSlabs(s, sco) for (s, sco) in zip(slabsco2, slabsco)]

    # %% Line-of-sight

    # Solve RTE along the line of sight
    # --------
    # two semi profile, + room absorption
    line_of_sight = slabstot[1:][::-1] + slabstot + [s0]
    stot = SerialSlabs(*line_of_sight)
    #        stot = SerialSlabs(*slabstot[1:][::-1], *slabstot, s0)  # Python 3 syntax only

    # Also calculate the contribution of pure CO
    # ---------
    s0tr = s0.copy()  # transmittance only
    s0tr.conditions["thermal_equilibrium"] = False
    s0tr._q["radiance_noslit"] *= 0  # hack to remove CO2 emission
    s0tr._q["emisscoeff"] *= 0  # hack to remove CO2 emission
    line_of_sight = slabsco[1:][::-1] + slabsco + [s0tr]
    sco = SerialSlabs(*line_of_sight)
    #        sco = SerialSlabs(*slabsco[1:][::-1], *slabsco, s0tr)  # Python 3 syntax only

    # Generate Slit
    # ------
    disp = 4  # dispersion conversion function (nm/mm)
    s_in = 1  # entrance slit (mm)
    s_out = 2.8  # exit slit (mm)
    top = (s_out - s_in) * disp
    base = (s_out + s_in) * disp
    slit_function = (top, base)  # (nm)

    # Final plot unit
    unit = "µW/cm2/sr"
    norm_by = "max"  # should be 'max' if unit ~ 'µW/cm2/sr',
    # 'area' if unit ~ 'µW/cm2/sr/nm'

    # Convolve with Slit
    # -------
    sco.apply_slit(slit_function, unit="nm", norm_by=norm_by, shape="trapezoidal")
    stot.apply_slit(slit_function, unit="nm", norm_by=norm_by, shape="trapezoidal")

    # Remove emission from within the spectrometer (subtracted
    # by the chopper experimentally)
    # -------
    s0spectro = s0.copy()
    s0spectro.rescale_path_length(75 * 4)  # spectrometer length
    s0spectro.apply_slit(slit_function, unit="nm", norm_by=norm_by, shape="trapezoidal")
    _, I0spectro = s0spectro.get("radiance", Iunit=unit)

    wtot, Itot = stot.get("radiance", wunit="nm", Iunit=unit)
    stot_corr = experimental_spectrum(
        wtot,
        Itot - I0spectro,  # hack to remove
        wunit="nm",  # in air by default
        # emission from within
        # spectrometer
        Iunit=unit,
    )

    # %% Compare with experiment data

    # Plot experimental data
    # ------
    exp = pd.read_csv(
        getValidationCase(
            "test_compare_torch_CO2_data/test_compare_torch_CO2_spectrum_JTHT2003.dat"
        ),
        skiprows=5,
        sep="\s+",
    )
    exp.w /= 10  # Angstrom to nm
    exp = exp[(exp.w > wlmin) & (exp.w < wlmax)]
    sexp = experimental_spectrum(exp.w, exp.I, wunit="nm", Iunit="mW/cm2/sr")

    if plot:

        sexp.plot("radiance", wunit="nm", Iunit=unit, lw=2, label="Packan 2003")

        # Plot calculated
        # ----------

        stot.plot(
            "radiance",
            wunit="nm",
            Iunit=unit,
            nfig="same",
            color="r",
            ls="--",
            label="All: CO$_2$+CO",
        )

        # Plot internal spectrometer emission
        s0spectro.plot(
            "radiance",
            wunit="nm",
            Iunit=unit,
            nfig="same",
            color="b",
            label="CO$_2$ in Spectrometer",
        )

        stot_corr.plot(
            "radiance",
            wunit="nm",
            Iunit=unit,
            nfig="same",
            color="r",
            lw=2,
            label="All-CO$_2$ in Spectro",
            zorder=10,
        )

        sco.plot(
            "radiance",
            wunit="nm",
            Iunit=unit,
            nfig="same",
            color="g",
            ls="-",
            label="CO",
        )

        plt.ylim((0, 9))
        plt.legend(loc="upper right")

        if save:
            plt.savefig("test_compare_torch_CO2_brd20.png")
            plt.savefig("test_compare_torch_CO2_brd20k.pdf")

    assert get_residual(stot_corr, sexp, "radiance") < 0.02

    if verbose:
        print(("Finished test_compare_torch_CO2 in {0:.0f}s".format(time() - t0)))


if __name__ == "__main__":
    printm("test_compare_torch_CO2:", test_compare_torch_CO2(plot=True, use_cache=True))



# -*- coding: utf-8 -*-
"""Print functions, with colors."""

import sys
from io import StringIO

import pandas as pd
from termcolor import colored

# %% Colored Print functions


def printm(*args, **kwargs):
    """Print in magenta.

    Typically use in RADIS for tests
    Could be used for Accuracy/Performance warnings."""

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color="magenta"))


def printg(*args, **kwargs):
    """Print in green.

    Typically use in RADIS for verbose>=2 messages"""

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color="green"))


def printr(*args, **kwargs):
    """Print in red.

    Typically use in RADIS to print errors that are corrected on-the-fly
    by the code"""

    string = _capture_print(*args, **kwargs)

    return print(colored(string, color="red"))


def _capture_print(*args, **kwargs):
    """Emulate print option but get output in a ```str``` variable instead of
    stdout."""

    # Change the output to capture the string instead of sending it to the console
    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()  # capture all print
    try:
        print(*args, **kwargs)
        string = newstdout.getvalue()  # Get string output
    except:
        raise
    finally:
        sys.stdout = old_stdout

    # discard line return character (will be added again when we print ``string`` )
    return string[:-1]


# %% Print in pandas


def print_full(x):
    """Print full Pandas series.

    From https://stackoverflow.com/questions/19124601/pretty-print-an-entire-pandas-series-dataframe
    """
    pd.set_option("display.max_rows", len(x))
    print(x)
    pd.reset_option("display.max_rows")


def get_print_full(x):
    """Same as print_full, but returns string."""

    # Change the output to capture the string instead of sending it to the console
    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()  # capture all print
    try:
        pd.set_option("display.max_rows", len(x))
        print(x)
        string = newstdout.getvalue()  # Get string output
    except:
        raise
    finally:
        sys.stdout = old_stdout
        pd.reset_option("display.max_rows")

    return string[:-1]


# %% UTF8 Upper script/ lower script
# https://stackoverflow.com/a/58612677/5622825

superscript_map = {
    "0": "⁰",
    "1": "¹",
    "2": "²",
    "3": "³",
    "4": "⁴",
    "5": "⁵",
    "6": "⁶",
    "7": "⁷",
    "8": "⁸",
    "9": "⁹",
    "a": "ᵃ",
    "b": "ᵇ",
    "c": "ᶜ",
    "d": "ᵈ",
    "e": "ᵉ",
    "f": "ᶠ",
    "g": "ᵍ",
    "h": "ʰ",
    "i": "ᶦ",
    "j": "ʲ",
    "k": "ᵏ",
    "l": "ˡ",
    "m": "ᵐ",
    "n": "ⁿ",
    "o": "ᵒ",
    "p": "ᵖ",
    "q": "۹",
    "r": "ʳ",
    "s": "ˢ",
    "t": "ᵗ",
    "u": "ᵘ",
    "v": "ᵛ",
    "w": "ʷ",
    "x": "ˣ",
    "y": "ʸ",
    "z": "ᶻ",
    "A": "ᴬ",
    "B": "ᴮ",
    "C": "ᶜ",
    "D": "ᴰ",
    "E": "ᴱ",
    "F": "ᶠ",
    "G": "ᴳ",
    "H": "ᴴ",
    "I": "ᴵ",
    "J": "ᴶ",
    "K": "ᴷ",
    "L": "ᴸ",
    "M": "ᴹ",
    "N": "ᴺ",
    "O": "ᴼ",
    "P": "ᴾ",
    "Q": "Q",
    "R": "ᴿ",
    "S": "ˢ",
    "T": "ᵀ",
    "U": "ᵁ",
    "V": "ⱽ",
    "W": "ᵂ",
    "X": "ˣ",
    "Y": "ʸ",
    "Z": "ᶻ",
    "+": "⁺",
    "-": "⁻",
    "=": "⁼",
    "(": "⁽",
    ")": "⁾",
}


subscript_map = {
    "0": "₀",
    "1": "₁",
    "2": "₂",
    "3": "₃",
    "4": "₄",
    "5": "₅",
    "6": "₆",
    "7": "₇",
    "8": "₈",
    "9": "₉",
    "a": "ₐ",
    "b": "♭",
    "c": "꜀",
    "d": "ᑯ",
    "e": "ₑ",
    "f": "բ",
    "g": "₉",
    "h": "ₕ",
    "i": "ᵢ",
    "j": "ⱼ",
    "k": "ₖ",
    "l": "ₗ",
    "m": "ₘ",
    "n": "ₙ",
    "o": "ₒ",
    "p": "ₚ",
    "q": "૧",
    "r": "ᵣ",
    "s": "ₛ",
    "t": "ₜ",
    "u": "ᵤ",
    "v": "ᵥ",
    "w": "w",
    "x": "ₓ",
    "y": "ᵧ",
    "z": "₂",
    "A": "ₐ",
    "B": "₈",
    "C": "C",
    "D": "D",
    "E": "ₑ",
    "F": "բ",
    "G": "G",
    "H": "ₕ",
    "I": "ᵢ",
    "J": "ⱼ",
    "K": "ₖ",
    "L": "ₗ",
    "M": "ₘ",
    "N": "ₙ",
    "O": "ₒ",
    "P": "ₚ",
    "Q": "Q",
    "R": "ᵣ",
    "S": "ₛ",
    "T": "ₜ",
    "U": "ᵤ",
    "V": "ᵥ",
    "W": "w",
    "X": "ₓ",
    "Y": "ᵧ",
    "Z": "Z",
    "+": "₊",
    "-": "₋",
    "=": "₌",
    "(": "₍",
    ")": "₎",
}
